#!/usr/bin/env python3
#
# This file is part of Script of Scripts (sos), a workflow system
# for the execution of commands and scripts in different languages.
# Please visit https://github.com/bpeng2000/SOS
#
# Copyright (C) 2016 Bo Peng (bpeng@mdanderson.org)
##
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
import os
import sys
import re
import copy
import atexit
import glob
import fnmatch
import keyword
import argparse
import textwrap
import traceback
import multiprocessing as mp
from collections import OrderedDict, defaultdict
from collections.abc import Sequence, Iterable
from itertools import tee, combinations

# Python 2.7 should also have this module
from io import StringIO

from . import __version__
from .actions import *
from .utils import env, Error, WorkflowDict, SoS_eval, SoS_exec, RuntimeInfo, \
    dehtml, getTermWidth, interpolate, shortRepr, glob_wildcards, apply_wildcards

class ArgumentError(Error):
    """Raised when an invalid argument is passed."""
    def __init__(self, msg):
        Error.__init__(self, msg)
        self.args = (msg, )

class ParsingError(Error):
    """Raised when a configuration file does not follow legal syntax."""

    def __init__(self, filename):
        Error.__init__(self, 'File contains parsing errors: %s' % filename)
        self.filename = filename
        self.errors = []
        self.args = (filename, )

    def append(self, lineno, line, msg):
        self.errors.append((lineno, line))
        self.message += '\n\t[line %2d]: %s\n%s' % (lineno, line, msg)


class StepInfo(object):
    '''A simple class to hold input, output, and index of step. Its attribute can
    only be set using an interface, and cannot be assigned. This is to make sure
    such information is not changed freely by malicious scripts'''
    def __init__(self):
        pass

    def set(self, key, value):
        object.__setattr__(self, key, value)

    def __setattr__(self, key, value):
        raise RuntimeError('Changing of step info {} is prohibited.'.format(key))

    def __repr__(self):
        return '{' + ', '.join('{}: {!r}'.format(x,y) for x,y in self.__dict__.items()) + '}'

def execute_step_process(step_process, global_process, locals, sigil, signature, workdir):
    '''A function that has a local dictionary (from SoS env.locals),
    a global process, and a step process. The processes are executed 
    in a separate process, independent of SoS. This makes it possible
    to execute the processes in background, on cluster, or submit as
    Celery tasks.'''
    env.register_process(os.getpid(), 'spawned_job with {} {}'
        .format(', '.join(locals['_input']), ', '.join(locals['_output'])))
    try:
        os.chdir(workdir)
        # switch context to the new dict and switch back once the with
        # statement ends (or if an exception is raised)
        with env.push_context(locals):
            if global_process:
                SoS_exec(global_process, sigil)
            SoS_exec(step_process, sigil)
        if signature:
            signature.write()
    except KeyboardInterrupt:
        raise RuntimeError('KeyboardInterrupt from {}'.format(os.getpid()))
    env.deregister_process(os.getpid())
    return 0

class SoS_Step:
    #
    # A single sos step
    #
    _INPUT_OPTIONS = ['group_by', 'skip', 'filetype', 'paired_with', 'for_each', 'pattern', 'dynamic']
    _OUTPUT_OPTIONS = ['pattern', 'dynamic']
    _DEPENDS_OPTIONS = ['pattern', 'dynamic']
    _RUNTIME_OPTIONS = ['workdir', 'concurrent']
    def __init__(self, names=[], options={}, is_global=False, is_parameters=False):
        # A step will not have a name and index until it is copied to separate workflows
        self.name = None
        self.index = None
        # it initially hold multiple names with/without wildcard characters
        self.names = names
        self.options = options
        self.comment = ''
        # comment at the end of a section that could be a workflow description
        self.back_comment = ''
        # parameters for parameters section
        self.parameters = []
        # everything before step process
        self.statements = []
        # step processes
        self.global_process = ''
        self.process = ''
        # subworkflow of a step
        self.subworkflow = None
        # is it global section? This is a temporary indicator because the global section 
        # will be inserted to each step of the workflow.
        self.is_global = is_global
        # is it the parameters section?
        self.is_parameters = is_parameters
        # indicate the type of input of the last line
        self.values = []
        self.lineno = None
        #
        self.runtime_options = {}
        if 'sigil' in self.options:
            self.sigil = self.options['sigil']
        else:
            self.sigil = '${ }'
        
        #
        # string mode to collect all strings as part of an action
        self._action = None
        self._script = ''

    def category(self):
        if self.statements:
            if self.statements[-1][0] == '=':
                return 'expression'
            elif self.statements[-1][0] == ':':
                # a hack. ... to avoid calling isValid recursively
                def validDirective():
                    if not self.values:
                        return True
                    if self.values[-1].strip().endswith(','):
                        return False
                    try:
                        compile('func(' + ''.join(self.values) + ')', filename='<string>', mode='eval')
                    except:
                        return False
                    return True
                if validDirective() and self._action is not None:
                    return 'script'
                else:
                    return 'directive'
            else:
                return 'statements'
        elif self.parameters:
            return 'expression'
        else:
            return None

    #
    # Parsing input
    #
    def empty(self):
        '''If there is no content (comment does not count)'''
        return self.category() is None

    def extend(self, line):
        if self.category() == 'directive':
            self.add_directive(None, line)
        elif self.category() == 'expression':
            self.add_assignment(None, line)
        elif self.category() == 'script':
            self._script += line
        else:
            self.add_statement(line)
        self.back_comment = ''

    def add_comment(self, line):
        '''Add comment line'''
        # in parameter section, comments will always be kept
        if self.is_parameters or self.empty():
            self.comment += line.lstrip('#').lstrip()
        else:
            self.back_comment += line.lstrip('#').lstrip()

    def add_assignment(self, key, value, lineno=None):
        '''Assignments are items with '=' type '''
        if key is None:
            # continuation of multi-line assignment
            if self.is_parameters:
                self.parameters[-1][1] += value
            else:
                self.statements[-1][-1] += value
            self.values.append(value)
        else:
            # new assignment
            if self.is_parameters:
                # in assignment section, comments belong to their following
                # parameter definition
                self.parameters.append([key, value, self.comment])
                self.comment = ''
            else:
                self.statements.append(['=', key, value])
            self.values = [value]
        self.back_comment = ''
        if lineno:
            self.lineno = lineno

    def add_directive(self, key, value, lineno=None, action=None):
        '''Assignments are items with ':' type '''
        if key is None:
            # continuation of multi-line directive
            self.statements[-1][-1] += value
            self.values.append(value)
        else:
            # new directive
            self.statements.append([':', key, value])
            self.values = [value]
            if action is not None:
                self._action = action
        self.back_comment = ''
        if lineno:
            self.lineno = lineno

    def add_statement(self, line, lineno=None):
        '''Assignments are items with ':' type '''
        # there can be only one statement block
        if self.category() != 'statements':
            self.values = [line]
        else:
            self.values.append(line)
        if self.statements and self.statements[-1][0] == '!':
            self.statements[-1][-1] += line
        else:
            self.statements.append(['!', line])
        self.back_comment = ''
        if lineno:
            self.lineno = lineno

    def wrap_script(self):
        '''convert action: script to process: action(script)'''
        if self._action is None:
            return
        self.statements.append(['!', '{}({!r})'.format(self._action, textwrap.dedent(self._script))])
        self._action = None
        self._script = ''

    def finalize(self):
        ''' split statement and process by last directive '''
        self.wrap_script()
        if not self.statements:
            self.process = ''
            return
        process_directive = [idx for idx, statement in enumerate(self.statements) if statement[0] == ':' and statement[1] == 'process']
        if not process_directive:
            self.process = ''
            return
        start_process = process_directive[0] + 1
        # convert statement to process
        self.process = ''
        for statement in self.statements[start_process:]:
            if statement[0] == '=':
                self.process += '{} = {}'.format(statement[1], statement[2])
            elif statement[0] == ':':
                if statement[1] in ('input', 'output', 'depends'):
                    raise ValueError('Step process should be defined as the last item in a SoS step')
                elif statement[2].strip():
                    raise ValueError('Runtime options are not allowed for second or more actions in a SoS step')
                # ignore ...
                self.process += '\n'
            else:
                self.process += statement[1]
        # remove process from self.statement
        if process_directive:
            self.statements = self.statements[:start_process]

    def isValid(self):
        if not self.values:
            return True
        try:
            if self.category() == 'expression':
                compile(''.join(self.values), filename='<string>', mode='eval')
            elif self.category() == 'directive':
                # we add func() because the expression can be multi-line and
                # can have keyword-argument like options
                #
                # However, python considers
                #
                #     func('value', )
                #
                # a valid syntax but we do want , to continue to the next line
                if self.values[-1].strip().endswith(','):
                    return False
                compile('func(' + ''.join(self.values) + ')', filename='<string>', mode='eval')
            elif self.category() == 'statements':
                compile(''.join(self.values), filename='<string>', mode='exec')
            elif self.category() == 'script':
                return True
            else:
                raise RuntimeError('Unrecognized expression type {}'.format(self.category()))
            return True
        except Exception as e:
            return False


    #
    # Execution: input options
    #
    def _handle_input_group_by(self, ifiles, group_by):
        # for this special case, the step is skipped
        if group_by == 'single':
            return [[x] for x in ifiles]
        elif group_by == 'all':
            # default option
            return [ifiles]
        elif group_by == 'pairs':
            if len(ifiles) % 2 != 0:
                raise ValueError('Paired group_by has to have even number of input files: {} provided'
                    .format(len(ifiles)))
            return list(list(x) for x in zip(ifiles[:len(ifiles)//2], ifiles[len(ifiles)//2:]))
        elif group_by == 'pairwise':
            f1, f2 = tee(ifiles)
            next(f2, None)
            return [list(x) for x in zip(f1, f2)]
        elif group_by == 'combinations':
            return [list(x) for x in combinations(ifiles, 2)]

    def _handle_input_pattern(self, pattern):
        ifiles = env.locals['_step'].input
        # 
        if pattern is None or not pattern:
            patterns = []
        elif isinstance(pattern, str):
            patterns = [pattern]
        elif isinstance(pattern, Iterable):
            patterns = pattern
        else:
            raise ValueError('Unacceptable value for parameter pattern: {}'.format(pattern))
        #
        for pattern in patterns:
            res = glob_wildcards(pattern, [])
            for ifile in ifiles:
                matched = glob_wildcards(pattern, [ifile])
                for key in matched.keys():
                    if not matched[key]:
                        env.logger.warning('Filename {} does not match pattern {}. None returned.'.format(ifile, pattern))
                        res[key].append(None)
                    else:
                        res[key].extend(matched[key])
            # now, assign the variables to env
            for k, v in res.items():
                env.locals[k] = v
            # also make k, v pair with _input
            self._handle_input_paired_with(res.keys())
        
    def _handle_input_paired_with(self, paired_with):
        if paired_with is None or not paired_with:
            paired_with = []
        elif isinstance(paired_with, str):
            paired_with = [paired_with]
        elif isinstance(paired_with, Iterable):
            paired_with = paired_with
        else:
            raise ValueError('Unacceptable value for parameter paired_with: {}'.format(paired_with))
        #
        ifiles = env.locals['_step'].input
        if not hasattr(self, '_vars'):
            self._vars = [{} for x in self._groups]
        for wv in paired_with:
            if '.' in wv:
                if wv.split('.')[0] not in env.locals:
                    raise ValueError('Variable {} does not exist.'.format(wv))
                values = getattr(env.locals[wv.split('.')[0]], wv.split('.', 1)[-1])
            else:
                if wv not in env.locals:
                    raise ValueError('Variable {} does not exist.'.format(wv))
                values = env.locals[wv]
            if isinstance(values, str) or not isinstance(values, Iterable):
                raise ValueError('with_var variable {} is not a sequence ("{}")'.format(wv, values))
            if len(values) != len(ifiles):
                raise ValueError('Length of variable {} (length {}) should match the number of input files (length {}).'
                    .format(wv, len(values), len(ifiles)))
            file_map = {x:y for x,y in zip(ifiles, values)}
            for idx, grp in enumerate(self._groups):
                self._vars[idx]['_' + wv.split('.')[0]] = [file_map[x] for x in grp]

    def _handle_input_for_each(self, for_each):
        if for_each is None or not for_each:
            for_each = []
        elif isinstance(for_each, str):
            for_each = [for_each]
        elif isinstance(for_each, Sequence):
            for_each = for_each
        else:
            raise ValueError('Unacceptable value for parameter for_each: {}'.format(for_each))
        #
        for fe_all in for_each:
            loop_size = None
            for fe in fe_all.split(','):
                values = env.locals[fe]
                if not isinstance(values, Sequence):
                    raise ValueError('for_each variable {} is not a sequence ("{}")'.format(fe, values))
                if loop_size is None:
                    loop_size = len(values)
                elif loop_size != len(values):
                    raise ValueError('Length of variable {} (length {}) should match the length of variable {} (length {}).'
                        .format(fe, len(values), fe_all.split(',')[0], loop_size))
            # expand
            self._groups = self._groups * loop_size
            tmp = []
            for vidx in range(loop_size):
                for idx in range(len(self._vars)):
                    for fe in fe_all.split(','):
                        if '.' in fe:
                            if fe.split('.')[0] not in env.locals:
                                raise ValueError('Variable {} does not exist.'.format(fe))
                            self._vars[idx]['_' + fe.split('.')[0]] = getattr(env.locals[fe.split('.')[0]], fe.split('.', 1)[-1])[vidx]
                        else:
                            if fe not in env.locals:
                                raise ValueError('Variable {} does not exist.'.format(fe))
                            self._vars[idx]['_' + fe] = env.locals[fe][vidx]
                tmp.extend(copy.deepcopy(self._vars))
            self._vars = tmp

    #
    # Execution; directives
    #
    def _directive_input(self, *args, **kwargs):
        # first *args are filenames
        for k in kwargs.keys():
            if k not in self._INPUT_OPTIONS:
                raise RuntimeError('Unrecognized input option {}'.format(k))
        #
        if args:
            ifiles = []
            for arg in args:
                if isinstance(arg, str):
                    ifiles.append(arg)
                elif isinstance(arg, list):
                    if not all(isinstance(x, str) for x in arg):
                        raise RuntimeError('Invalid input file: {}'.format(arg))
                    ifiles.extend(arg)
            env.locals['_step'].set('input', ifiles)
        else:
            ifiles = env.locals['_step'].input
        # expand files with wildcard characters and check if files exist
        tmp = []
        for ifile in ifiles:
            if os.path.isfile(os.path.expanduser(ifile)):
                tmp.append(ifile)
            elif env.run_mode == 'run':
                # in this mode file must exist
                expanded = glob.glob(os.path.expanduser(ifile))
                if not expanded:
                    raise RuntimeError('{} not exist.'.format(ifile))
                tmp.extend(expanded)
            elif env.run_mode == 'dryrun':
                # FIXME: this should be the 'dynamic' mode
                expanded = glob.glob(os.path.expanduser(ifile))
                if expanded:
                    tmp.extend(expanded)
                else:
                    tmp.append(ifile)
        #
        ifiles = tmp
        #
        if 'filetype' in kwargs:
            if isinstance(kwargs['filetype'], str):
                ifiles = fnmatch.filter(ifiles, kwargs['filetype'])
            elif isinstance(kwargs['filetype'], (list, tuple)):
                ifiles = [x for x in ifiles if any(fnmatch.fnmatch(x, y) for y in kwargs['filetype'])]
            elif callable(kwargs['filetype']):
                ifiles = [x for x in ifiles if kwargs['filetype'](x)]
        #             
        # handle group_by
        if 'group_by' in kwargs:
            self._groups = self._handle_input_group_by(ifiles, kwargs['group_by'])
        else:
            self._groups = [ifiles]
        #
        self._vars = [{} for x in self._groups]
        # handle paired_with
        if 'paired_with' in kwargs:
            self._handle_input_paired_with(kwargs['paired_with'])
        # handle pattern
        if 'pattern' in kwargs:
            self._handle_input_pattern(kwargs['pattern'])
        # handle for_each
        if 'for_each' in kwargs:
            self._handle_input_for_each(kwargs['for_each'])
        #
        if 'skip' in kwargs:
            if callable(kwargs['skip']):
                # FIXME:
                # this is a quick fix to make sure user-defined functions are avaialble to use
                # by another user-defined funciton. There should be better methods out there.
                globals().update({x:y for x,y in env.locals.items() if callable(y)})
                _groups = []
                _vars = []
                for g, v in zip(self._groups, self._vars):
                    try:
                        res = kwargs['skip'](g, **v)
                    except Exception as e:
                        raise RuntimeError('Failed to apply skip function to group {}: {}'.format(', '.join(g), e))
                    if res:
                       env.logger.info('Input group {} is skipped.'.format(', '.join(g)))
                    else:
                        _groups.append(g)
                        _vars.append(v)
                self._groups = _groups
                self._vars = _vars
            else:
                self._groups = []
                self._vars = []


    def _directive_depends(self, *args, **kwargs):
        '''handle directive depends'''
        for k in kwargs.keys():
            if k not in self._DEPENDS_OPTIONS:
                raise RuntimeError('Unrecognized depends option {}'.format(k))
        # first *args are filenames
        dfiles = []
        for arg in args:
            if isinstance(arg, str):
                dfiles.append(arg)
            elif isinstance(arg, list):
                if not all(isinstance(x, str) for x in arg):
                    raise RuntimeError('Invalid dependent file: {}'.format(arg))
                dfiles.extend(arg)
        env.locals.set('_depends', dfiles)

    def _handle_output_pattern(self, pattern, ofiles):
        # 
        if pattern is None or not pattern:
            patterns = []
        elif isinstance(pattern, str):
            patterns = [pattern]
        elif isinstance(pattern, list):
            patterns = pattern
        else:
            raise ValueError('Unacceptable value for parameter pattern: {}'.format(pattern))
        #
        for pattern in patterns:
            sz = None
            res = glob_wildcards(pattern, [])
            sz = None
            wildcard = [{}]
            for key in res.keys():
                if key not in env.locals:
                    raise ValueError('Undefined variable {} in pattern {}'.format(key, pattern))
                if not isinstance(env.locals[key], str) and isinstance(env.locals[key], Sequence):
                    if sz is None:
                        sz = len(env.locals[key])
                        wildcard = [{} for x in range(sz)]
                    elif sz != len(env.locals[key]):
                        raise ValueError('Variables in output pattern should have the same length (other={}, len({})={})'
                            .format(sz, key, len(env.locals[key])))
                    for idx, value in enumerate(env.locals[key]):
                        wildcard[idx][key] = value
                else:
                    for v in wildcard:
                        v[key] = env.locals[key]
            #
            for card in wildcard:
                ofiles.append(apply_wildcards(pattern, card, fill_missing=False,
                   fail_dynamic=False, dynamic_fill=None, keep_dynamic=False))
            
    def _directive_output(self, *args, **kwargs):
        for k in kwargs.keys():
            if k not in self._OUTPUT_OPTIONS:
                raise RuntimeError('Unrecognized output option {}'.format(k))
        ofiles = []
        for arg in args:
            if isinstance(arg, str):
                ofiles.append(arg)
            elif isinstance(arg, list):
                if not all(isinstance(x, str) for x in arg):
                    raise RuntimeError('Invalid output file: {}'.format(arg))
                ofiles.extend(arg)
        #
        if 'pattern' in kwargs:
            self._handle_output_pattern(kwargs['pattern'], ofiles)
        env.locals.set('_output', ofiles)

    def _directive_process(self, **kwargs):
        for k in kwargs.keys():
            if k not in self._RUNTIME_OPTIONS:
                raise RuntimeError('Unrecognized runtime option {}'.format(k))
        #
        self.runtime_options = kwargs

    #
    # handling of parameters step
    #
    def _parse_error(self, msg):
        '''This function will replace error() function in argparse module so that SoS
        can hijack errors raised from it.'''
        raise ArgumentError(msg)

    def parse_args(self, args, check_unused=False, cmd_name=''):
        '''Parse command line arguments and set values to parameters section'''
        env.logger.info('Execute ``{}_parameters``'.format(self.name))
        env.locals['_step'].set('name', '{}_parameters'.format(self.name))
        env.locals['_step'].set('index', -1)
        if self.global_process:
            try:
                SoS_exec(self.global_process)
            except Exception as e:
                if env.verbosity > 2:
                    print_traceback()
                raise RuntimeError('Failed to execute statements\n"{}"\n{}'.format(self.global_process, e))
     
        def str2bool(v):
            if v.lower() in ('yes', 'true', 't', '1'):
                return True
            elif v.lower() in ('no', 'false', 'f', '0'):
                return False
            else:
                raise ArgumentError('Invalid value for bool argument "{}" (only yes,no,true,false,t,f,0,1 are allowed)'.format(v))
        #
        parser = argparse.ArgumentParser(prog='sos-runner {}'.format(cmd_name))
        parser.register('type', 'bool', str2bool)
        arguments = {}
        for key, defvalue, comment in self.parameters:
            try:
                defvalue = SoS_eval(defvalue, self.sigil)
                arguments[key] = defvalue
            except Exception as e:
                raise RuntimeError('Incorrect default value {} for parameter {}: {}'.format(defvalue, key, e))
            if isinstance(defvalue, type):
                if defvalue == bool:
                    parser.add_argument('--{}'.format(key), type='bool', help=comment, required=True, nargs='?') 
                else:
                    # if only a type is specified, it is a required document of required type
                    parser.add_argument('--{}'.format(key), type=str if hasattr(defvalue, '__iter__') else defvalue,
                        help=comment, required=True, nargs='+' if hasattr(defvalue, '__iter__') else '?')
            else:
                if isinstance(defvalue, bool):
                    parser.add_argument('--{}'.format(key), type='bool', help=comment,
                        nargs='?', default=defvalue)
                else:
                    parser.add_argument('--{}'.format(key), type=type(defvalue), help=comment,
                        nargs='*' if isinstance(defvalue, Sequence) and not isinstance(defvalue, str) else '?',
                        default=defvalue)
        #
        parser.error = self._parse_error
        # 
        # because of the complexity of having combined and nested workflows, we cannot know how
        # many parameters section a workflow has and therfore have to assume that the unknown parameters
        # are for other sections.
        if check_unused:
            parsed = parser.parse_args(args)
        else:
            parsed, unknown = parser.parse_known_args(args)
            if unknown:
                env.logger.warning('Unparsed arguments [{}] that might be processed by another combined or nested workflow'
                    .format(' '.join(unknown)))
        #
        arguments.update(vars(parsed))
        # now change the value with passed values
        for k, v in arguments.items():
            env.locals[k] = v
            # protect variables from being further modified
            env.readonly_vars.add(k)

    #
    # Execution
    #
    def step_signature(self):
        '''return everything that might affect the execution of the step 
        namely, global process, step definition etc to create a unique 
        signature that might will be changed with the change of SoS script.'''
        result = self.global_process
        for statement in self.statements:
            if statement[0] in (':', '='):
                result += '{}: {}\n'.format(statement[1], statement[2])
            else:
                result += statement[1] + '\n'
        result += self.process
        return re.sub(r'\s+', ' ', result)

    def run(self, pool):
        # handle these two sections differently
        if isinstance(self.index, int):
            env.locals.set('_workflow_index', self.index)
        #
        env.logger.info('Execute ``{}_{}``: {}'.format(self.name, self.index, self.comment.strip()))
        #
        # 
        # the following is a quick hack to allow _directive_input function etc to access 
        # the workflow dictionary
        env.globals['self'] = self
        env.locals['_step'].set('name', '{}_{}'.format(self.name, self.index))
        env.locals['_step'].set('index', self.index)
        env.locals['_step'].set('output', [])
        env.locals['_step'].set('depends', [])
        #
        # these are temporary variables that should be removed if exist
        for var in ('_input', '_depends', '_output'):
            env.locals.pop(var, '')
        #
        # default input groups and vars
        if hasattr(env.locals['_step'], 'input'):
            self._groups = [env.locals['_step'].input]
        else:
            self._groups = [[]]
        self._vars = [{}]
        #
        input_idx = [idx for idx,x in enumerate(self.statements) if x[0] == ':' and x[1] == 'input']
        if not input_idx:
            input_idx = None
        elif len(input_idx) == 1:
            input_idx = input_idx[0]
        else:
            raise RuntimeError('Only one step input is allowed')
        # if there is no input, easily
        self._outputs = []
        self._depends = []
        #
        # execute global process
        if self.global_process:
            try:
                SoS_exec(self.global_process)
            except Exception as e:
                if env.verbosity > 2:
                    print_traceback()
                raise RuntimeError('Failed to execute statements\n"{}"\n{}'.format(self.global_process, e))
        #
        if input_idx is not None:
            # execute before input stuff
            for statement in self.statements[:input_idx]:
                if statement[0] == '=':
                    key, value = statement[1:]
                    try:
                        env.locals[key] = SoS_eval(value, self.sigil)
                    except Exception as e:
                        raise RuntimeError('Failed to assign {} to variable {}: {}'.format(value, key, e))
                elif statement[0] == ':':
                    raise RuntimeError('Step input should be specified before others')
                else:
                    try:
                        SoS_exec(statement[1], self.sigil)
                    except Exception as e:
                        raise RuntimeError('Failed to process statement {}: {}'.format(statement[1], e))
            # input
            key, value = self.statements[input_idx][1:]
            try:
                SoS_eval('self._directive_{}({})'.format(key, value), self.sigil)
            except Exception as e:
                raise RuntimeError('Failed to process directive {} ({}): {}'.format(key, value, e))
            input_idx += 1
        else:
            # assuming everything starts from 0 is after input
            input_idx = 0
        
        if 'alias' in self.options:
            # copy once before the step might be skipped...
            env.locals[self.options['alias']] = copy.deepcopy(env.locals['_step'])
        if not self._groups:
            env.logger.info('Step {} is skipped'.format(self.index))
            return
        # post input
        # output and depends can be processed many times
        step_sig = self.step_signature()
        env.logger.info('_step.input:   ``{}``'.format(shortRepr(env.locals['_step'].input)))
        for idx, (g, v) in enumerate(zip(self._groups, self._vars)):
            # other variables
            env.locals.update(v)
            env.locals.set('_input', g)
            env.locals.set('_index', idx)
            for key in ('_output', '_depends'):
                if key in env.locals:
                    env.locals.pop(key)
            for statement in self.statements[input_idx:]:
                if statement[0] == '=':
                    key, value = statement[1:]
                    try:
                        env.locals[key] = SoS_eval(value, self.sigil)
                    except Exception as e:
                        raise RuntimeError('Failed to assign {} to variable {}: {}'.format(value, key, e))
                elif statement[0] == ':':
                    key, value = statement[1:]
                    # input and runtime is processed only once
                    try:
                        SoS_eval('self._directive_{}({})'.format(key, value), self.sigil)
                    except Exception as e:
                        raise RuntimeError('Failed to process directive {}: {}'.format(key, e))
                else:
                    old_run_mode = env.run_mode
                    if '_output' in env.locals:
                        signature = RuntimeInfo(step_sig, env.locals['_input'], env.locals['_output'],
                            env.locals.get('_depends', []))
                        if env.sig_mode == 'default' and signature.validate():
                            env.logger.info('Execute statement in dryrun mode and reuse existing output files {}'.format(', '.join(env.locals['_output'])))
                            env.run_mode = 'dryrun'
                    try:
                        SoS_exec(statement[1], self.sigil)
                    except Exception as e:
                        raise RuntimeError('Failed to process statement {}: {}'.format(statement[1], e))
                    env.run_mode = old_run_mode

            # collect _output and _depends
            if '_output' in env.locals:
                self._outputs.append(env.locals['_output'])
            else:
                self._outputs.append([])
            if '_depends' in env.locals:
                self._depends.append(env.locals['_depends'])
            else:
                self._depends.append([])
        #
        # if no output directive, assuming no output for each step
        if not self._outputs:
            self._outputs = [[] for x in self._groups]
        # if no depends directive, assuming no dependent files for each step
        if not self._depends:
            self._depends = [[] for x in self._groups]
        # we need to reduce output files in case they have been processed multiple times.
        env.locals['_step'].set('output', list(OrderedDict.fromkeys(sum(self._outputs, []))))
        env.locals['_step'].set('depends', list(OrderedDict.fromkeys(sum(self._depends, []))))
        env.logger.info('_step.output:  ``{}``'.format(shortRepr(env.locals['_step'].output)))
        if env.locals['_step'].depends:
            env.logger.info('_step.depends: ``{}``'.format(shortRepr(env.locals['_step'].depends)))
        if 'alias' in self.options:
            env.locals[self.options['alias']] = copy.deepcopy(env.locals['_step'])
        #
        # if the signature matches, the whole step is ignored, including subworkflows
        if env.locals['_step'].output:
            signature = RuntimeInfo(step_sig, 
                env.locals['_step'].input, env.locals['_step'].output, env.locals['_step'].depends)
            if env.run_mode == 'run':
                for ofile in env.locals['_step'].output:
                    parent_dir = os.path.split(os.path.expanduser(ofile))[0]
                    if parent_dir and not os.path.isdir(parent_dir):
                        os.makedirs(parent_dir)
                if env.sig_mode == 'default':
                    if signature.validate():
                        # everything matches
                        env.logger.info('Reusing existing output files {}'.format(', '.join(env.locals['_step'].output)))
                        return
                elif env.sig_mode == 'assert':
                    if not signature.validate():
                        raise RuntimeError('Signature mismatch.')
                # elif env.sig_mode == 'ignore'
        else:
            signature = None
        #
        results = []
        signatures = []
        for idx, (g, v, o, d) in enumerate(zip(self._groups, self._vars, self._outputs, self._depends)):
            env.locals.update(v)
            env.locals.set('_input', g)
            env.locals.set('_output', o)
            env.locals.set('_depends', d)
            env.locals.set('_index', idx)
            env.logger.info('_idx: ``{}``'.format(idx))
            env.logger.info('_input: ``{}``'.format(shortRepr(env.locals['_input'])))
            env.logger.info('_output: ``{}``'.format(shortRepr(env.locals['_output'])))
            #
            # action
            # If the users specifies output files for each loop (using ${input} etc, we
            # can try to see if we can create partial signature. This would help if the
            # step is interrupted in the middle.
            partial_signature = None
            if env.locals['_output'] and env.locals['_output'] != env.locals['_step'].output and env.run_mode == 'run':
                partial_signature = RuntimeInfo(step_sig, env.locals['_input'], env.locals['_output'], 
                    env.locals['_depends'])
                if env.sig_mode == 'default':
                    if partial_signature.validate():
                        # everything matches
                        env.logger.info('Reusing existing output files {}'.format(', '.join(env.locals['_output'])))
                        continue
                elif env.sig_mode == 'assert':
                    if not partial_signature.validate():
                        raise RuntimeError('Signature mismatch for input {} and output {}'.format(
                            ', '.join(env.locals['_input']), ', '.join(env.locals['_output'])))
            # now, if output file has already been generated using non-process statement
            # so that no process need to be run, or if the output need help from a subworkflow
            # we create signature from outside.
            if (not self.process or self.subworkflow) and partial_signature is not None:
                signatures.append(partial_signature)
            #
            try:
                # in case of nested workflow, these names might be changed and need to be reset
                env.locals['_step'].set('name', '{}_{}'.format(self.name, self.index))
                env.locals['_step'].set('index', self.index)
                #
                if self.process:
                    if 'concurrent' not in self.runtime_options or self.runtime_options['concurrent']:
                        submitter = pool.apply_async
                    else:
                        submitter = pool.apply
                    results.append(submitter(
                        execute_step_process,   # function
                        (self.process,          # process
                        self.global_process,    # global process
                        env.locals.clone_pickleable(),
                        self.sigil,
                        # if subworkflow contribute to outcome, we can not save signature here
                        partial_signature if partial_signature is not None and not self.subworkflow else None,
                        self.runtime_options['workdir'] if 'workdir' in self.runtime_options else os.getcwd())))
            except Exception as e:
                # FIXME: cannot catch exception from subprocesses
                if env.verbosity > 2:
                    print_traceback()
                raise RuntimeError('Failed to execute process\n"{}"\n{}'.format(self.process, e))
                #
            try:
                # if there is subworkflow, the subworkflow is considered part of the process
                # and will be executed mutliple times if necessary. The partial signature 
                # stuff works for these though
                if self.subworkflow:
                    self.subworkflow.run(nested=True, existing_pool=pool)
            except Exception as e:
                # FIXME: cannot catch exception from subprocesses
                if env.verbosity > 2:
                    print_traceback()
                raise RuntimeError('Failed to execute subworkflow: {}'.format(e))
        # check results?
        try:
            results = [res.get() if isinstance(res, mp.pool.AsyncResult) else res for res in results]
        except KeyboardInterrupt:
            # if keyboard interrupt
            pool.terminate()
            pool.join()
            raise RuntimeError('KeyboardInterrupt fro m {} (master)'.format(os.getpid()))
        except Exception as e:
            # if keyboard interrupt etc
            env.logger.error('Caught {}'.format(e))
            pool.terminate()
            pool.join()
            raise
        for sig in signatures:
            sig.write()
        if not all(x==0 for x in results):
            raise RuntimeError('Step process returns non-zero value')
        if env.run_mode == 'run':
            for ofile in env.locals['_step'].output:
                if not os.path.isfile(os.path.expanduser(ofile)): 
                    raise RuntimeError('Output file {} does not exist after completion of action'.format(ofile))
        if signature and env.run_mode == 'run':
            signature.write()

    def show(self, indent):
        '''Output for command sos show'''
        textWidth = max(60, getTermWidth())
        if self.is_parameters:
            print(indent + '  Parameters:')
            for k,v,c in self.parameters:
                 text = '{:<18}'.format(k) + (c + ' ' if c else '') + \
                     ('(default: {})'.format(v) if v else '')
                 print('\n'.join(
                     textwrap.wrap(text, 
                     initial_indent=indent + ' '*4,
                     subsequent_indent=indent + ' '*22,
                     width=textWidth)
                     ))
        else:
            # hide a step if there is no comment
            text = '  {:<20}'.format('Step {}_{}:'.format(self.name, self.index)) + self.comment
            print('\n'.join(
                textwrap.wrap(text, 
                    width=textWidth, 
                    initial_indent=indent,
                    subsequent_indent=indent + ' '*22)))
            if self.subworkflow:
                self.subworkflow.show(indent+ '   ' + chr(124), nested=True)
          

class SoS_Workflow:
    #
    # A SoS workflow with multiple steps
    #
    def __init__(self, workflow_name, allowed_steps, sections, description):
        '''create a workflow from its name and a list of SoS_Sections (using name matching)'''
        self.name = workflow_name
        self.description = description
        self.sections = []
        self.auxillary_sections = []
        #
        for section in sections:
            if section.is_parameters:
                self.sections.append(copy.deepcopy(section))
                self.sections[-1].name = workflow_name
                # for ordering purpose, this section is always after global
                self.sections[-1].index = -1
                # parameters and global section will not have subworkflow
                self.sections[-1].subworkflow = None
                continue
            for name, index, subworkflow in section.names:
                if 'target' in section.options:
                    self.auxillary_sections.append(section)
                elif fnmatch.fnmatch(workflow_name, name):
                    self.sections.append(copy.deepcopy(section))
                    self.sections[-1].name = workflow_name
                    self.sections[-1].index = 0 if index is None else int(index)
                    self.sections[-1].subworkflow = subworkflow
        #
        # sort sections by index
        self.sections.sort(key=lambda x: x.index)
        #
        # disable some disallowed steps
        if allowed_steps:
            all_steps = {x.index:False for x in self.sections if x.index >= 0}
            #
            for item in allowed_steps.split(','):
                # remove space
                item = ''.join([x for x in item if x != ' '])
                if item.isdigit():
                    # pipeline:100
                    all_steps[int(item)] = True
                elif '-' in item and item.count('-') == 1:
                    l, u = item.split('-')
                    if (l and not l.isdigit()) or (u and not u.isdigit()) or \
                        (l and u and int(l) > int(u)):
                        raise ValueError('Invalid pipeline step item {}'.format(item))
                    # pipeline:-100, pipeline:100+ or pipeline:10-100
                    if not l:
                        l = min(all_steps.keys())
                    if not u:
                        u = max(all_steps.keys())
                    #
                    for key in all_steps.keys():
                        if key >= int(l) and key <= int(u):
                            all_steps[key] = True
                else:
                    raise ValueError('Invalid pipeline step item {}'.format(item))
            # keep only selected steps (and the global and parameters section)
            self.sections = [x for x in self.sections if x.index < 0 or all_steps[x.index]]
        #
        for section in self.sections:
            if section.subworkflow is not None and isinstance(section.subworkflow, str):
                raise RuntimeError('Subworkflow {} not executable most likely because of recursice expansion.'.format(section.subworkflow))
        #
        env.logger.debug('Workflow {} created with {} sections: {}'
            .format(workflow_name, len(self.sections),
            ', '.join('{}_{}'.format(section.name, 
                    'global' if section.index == -2 else ('parameters' if section.index == -1 else section.index))
            for section in self.sections)))

    def extend(self, workflow):
        '''Append another workflow to existing one to created a combined workflow'''
        # all sections are simply appended ...
        self.sections.extend(workflow.sections)

    def run(self, args=[], nested=False, cmd_name='', existing_pool=None):
        '''Execute a workflow with specified command line args. If sub is True, this 
        workflow is a nested workflow and be treated slightly differently.
        '''
        if nested:
            # if this is a subworkflow, we use _input as step input the workflow
            env.locals['_step'].set('input', env.locals['_input'])
        else:
            # other wise we need to prepare running environment.
            env.globals = globals()
            # Because this workflow might belong to a combined workflow, we do not clear
            # locals before the execution of workflow.
            #if not hasattr(env, 'locals'):
            env.locals = WorkflowDict()
            # initial values
            env.locals.set('SOS_VERSION', __version__)
            py_version = sys.version_info
            #
            env.locals.set('_step', StepInfo())
            # initially there is no input, output, or depends
            env.locals['_step'].set('input', [])
            env.locals['_step'].set('name', self.name)
            env.locals['_step'].set('output', [])
            env.locals['_step'].set('depends', [])
        #
        # process step of the pipelinp
        #
        start = True
        num_parameters_sections = len([x for x in self.sections if x.is_parameters or x.subworkflow])
        if num_parameters_sections == 0 and args:
            raise ArgumentError('Unused parameter {}'.format(' '.join(args)))
        #
        if existing_pool is None:
            pool = mp.Pool(env.max_jobs)
        else:
            pool = existing_pool
        # the steps can be executed in the pool (Not implemented)
        for idx, section in enumerate(self.sections):
            # global section will not change _step etc
            if section.is_parameters:
                # if there is only one parameters section and no nested workflow, check unused section
                section.parse_args(args, num_parameters_sections == 1, cmd_name=cmd_name)
                continue
            # 
            # execute section with specified input
            # 1. for first step of workflow, _step.input=[]
            # 2. for subworkflow, _step.input = _input
            # 3. for second to later step, _step.input = _step.output
            if not start:
                # set the output to the input of the next step
                if hasattr(env.locals['_step'], 'output'):
                    # passing step output to _step.input of next step
                    env.locals['_step'].set('input', env.locals['_step'].output)
                    # step_output and depends are temporary
                else:
                    env.locals['_step'].set('input', [])
                #
                env.locals['_step'].set('output', [])
                env.locals['_step'].set('depends', [])
            else:
                # use initial values
                start = False
            # each section can use multiple processes of the pool
            section.run(pool)
        if existing_pool is None:
            pool.close()
            pool.join()

    def show(self, indent = '', nested=False):
        textWidth = max(60, getTermWidth())
        paragraphs = dehtml(self.description).split('\n\n')
        print('\n'.join(
            textwrap.wrap('{} {}:  {}'.format(
                'Nested workflow' if nested else 'Workflow', 
                self.name, paragraphs[0]),
                initial_indent = indent,
                subsequent_indent = indent,
                width=textWidth)
            ))
        for paragraph in paragraphs[1:]:
            print('\n'.join(
            textwrap.wrap(paragraph, width=textWidth,
                initial_indent = indent,
                subsequent_indent = indent)
            ))
        for section in self.sections:
            section.show(indent)

class SoS_Script:
    _DIRECTIVES = ['input', 'output', 'depends', 'process']
    _SECTION_OPTIONS = ['alias', 'nonconcurrent',
        'skip', 'blocking', 'sigil', 'target', 'source']
    _PARAMETERS_SECTION_NAME = 'parameters'

    # Regular expressions for parsing section headers and options
    _SECTION_HEADER_TMPL = r'''
        ^\[\s*                             # [
        (?P<section_name>[\d\w_,+=*\s-]+)  # digit, alphabet, _ and ,
        (:\s*                              # :
        (?P<section_option>.*)             # section options
        )?                                 # optional
        \]\s*$                             # ]
        '''

    _SECTION_NAME_TMPL = '''
        ^\s*                               # start
        (?P<name>                          # optional name
        [a-zA-Z*]                          # alphabet or '*'
        ([\w\d_*]*?                        # followed by alpha numeric or '*'
        [a-zA-Z\d*])??                     # but last character cannot be _
        )?                                 # name is optional
        (?(name)                           # if there is name
        (_(?P<index>\d+))?                 #   optional _index
        |(?P<default_index>\d+))           # no name, then index
        (\s*=\s*
        (?P<subworkflow>                   # = subworkflow
        [\w\d_+\s-]+                       # subworkflow specification
        ))?                                # optional
        \s*$
        '''

    _SUBWORKFLOW_TMPL = '''
        ^\s*                               # leading space
        (?P<name>                          # name
        [a-zA-Z*]                          # cannot start with _ etc
        ([\w\d_]*?))                       # can have _ and digit
        (_(?P<steps>                       # index start from _
        [\d\s-]+))?                        # with - and digit
        \s*$                               # end
        '''

    _SECTION_OPTION_TMPL = '''
        ^\s*                               # start
        (?P<name>{})                       # one of the option names
        (\s*=\s*                           # =
        (?P<value>.+)                      # value
        )?                                 # value is optional
        \s*$
        '''.format('|'.join(_SECTION_OPTIONS))


    _FORMAT_LINE_TMPL = r'''
        ^                                  # from first column
        \#fileformat\s*=\s*                # starts with #fileformat=SOS
        (?P<format_name>.*)                # format name
        \s*$                               # till end of line
        '''

    _FORMAT_VERSION_TMPL = r'''
        ^                                  # from first column
        (?P<format_name>[a-zA-Z]+)         # format name
        (?P<format_version>[\d\.]+)        # any number and .
        \s*$                               # till end of line
        '''

    _DIRECTIVE_TMPL = r'''
        ^                                  # from start of line
        (?P<directive_name>                # 
        (?!({})\s*:)                       # not a python keyword followed by : (can be input)
        ({}                                # name of directive
        |[a-zA-Z][\w\d_]*))                #    or action
        \s*:\s*                            # followed by :
        (?P<directive_value>.*)            # and values
        '''.format('|'.join(keyword.kwlist), '|'.join(_DIRECTIVES))

    _ASSIGNMENT_TMPL = r'''
        ^                                  # from start of line
        (?P<var_name>[\w_][\d\w_]*)        # variable name
        \s*=\s*                            # assignment
        (?P<var_value>.*)                  # variable content
        '''

    SECTION_HEADER = re.compile(_SECTION_HEADER_TMPL, re.VERBOSE)
    SECTION_NAME = re.compile(_SECTION_NAME_TMPL, re.VERBOSE)
    SUBWORKFLOW = re.compile(_SUBWORKFLOW_TMPL, re.VERBOSE)
    SECTION_OPTION = re.compile(_SECTION_OPTION_TMPL, re.VERBOSE)
    FORMAT_LINE = re.compile(_FORMAT_LINE_TMPL, re.VERBOSE)
    FORMAT_VERSION = re.compile(_FORMAT_VERSION_TMPL, re.VERBOSE)
    DIRECTIVE = re.compile(_DIRECTIVE_TMPL, re.VERBOSE)
    ASSIGNMENT = re.compile(_ASSIGNMENT_TMPL, re.VERBOSE)

    def __init__(self, content='', filename=None):
        '''Parse a sectioned SoS script file. Please refer to the SoS manual
        for detailed specification of this format.

        Parameter `content` can be either a filename or a content of a
        SoS script in unicode, which is convenient for passing scripts for
        testing purposes.

        Parameter `filename` should be used if the content should be read 
        from a file.
        '''
        if filename:
            self.sos_script = os.path.expanduser(filename)
            if not os.path.isfile(self.sos_script):
                raise ValueError('{} does not exist'.format(filename))
            with open(self.sos_script) as fp:
                self._read(fp)
        else:
            self.sos_script = '<string>'
            with StringIO(content) as fp:
                self._read(fp)
        #
        # workflows in this script, from sections that are not skipped.
        section_steps = sum([x.names for x in self.sections if \
            not (x.is_global or x.is_parameters) and \
            not ('skip' in x.options and (x.options['skip'] is None or x.options['skip'])) and \
            not ('target' in x.options)], [])
        # (name, None) is auxiliary steps
        self.workflows = list(set([x[0] for x in section_steps if '*' not in x[0]]))
        if not self.workflows:
            self.workflows = ['default']
        #
        # now we need to record the workflows to the global and parameters section so 
        # that we know if which has to be included when a subworkflow is used.
        for section in self.sections:
            if section.is_global or section.is_parameters:
                section.names = self.workflows
        #
        # get script descriptions
        cur_description = None
        self.description = ''
        self.workflow_descriptions = defaultdict(str)
        for block in self.descriptions:
            lines = [x for x in block.split('\n') if x.strip()]
            if not lines:
                continue
            for name in self.workflows:
                if lines[0].strip() == name:
                    cur_description = name
                    break
            if cur_description:
                self.workflow_descriptions[cur_description] += '\n'.join(lines[1:] if lines[0].strip() == cur_description else lines) + '\n'
            else:
                self.description += block + '\n'
        for section in self.sections:
            lines = [x for x in section.back_comment.split('\n') if x.strip()]
            for name in self.workflows:
                if lines and lines[0].strip() == name:
                    self.workflow_descriptions[name] += '\n'.join(lines[1:]) + '\n'


    def _read(self, fp):
        self.sections = []
        self.format_version = '1.0'
        self.descriptions = []
        #
        comment_block = 1
        # cursect always point to the last section
        cursect = None
        last_expression = []
        last_statement = []
        all_step_names = []
        #
        # this ParsingError is a container for all parsing errors. It will be
        # raised after parsing if there is at least one parsing error.
        parsing_errors = ParsingError(self.sos_script)
        for lineno, line in enumerate(fp, start=1):
            #
            # comments in SoS scripts are mostly informative
            if line.startswith('#'):
                # Comment blocks before any section
                if cursect is None:
                    if comment_block == 1:
                        # look for format information
                        mo = self.FORMAT_LINE.match(line)
                        if mo:
                            format_name = mo.group('format_name')
                            if not format_name.upper().startswith('SOS'):
                                parsing_errors.append(lineno, line,
                                    'Unrecognized file format name {}. Expecting SOS.'.format(format_name))
                            mo = self.FORMAT_VERSION.match(format_name)
                            if mo:
                                self.format_version = mo.group('format_version')
                            else:
                                parsing_errors.append(lineno, line,
                                    'Unrecognized file format version in {}.'.format(format_name))
                    elif comment_block > 1:
                        # anything before the first section can be pipeline
                        # description.
                        self.descriptions[-1] += line.lstrip('#').lstrip()
                else:
                    # in the parameter section, the comments are description
                    # of parameters and are all significant
                    if cursect.isValid():
                        cursect.add_comment(line)
                    # comment add to script
                    else:
                        cursect.extend(line)
                continue
            elif not line.strip():
                # a blank line start a new comment block if we are still
                # in the front of the script
                if cursect is None:
                    comment_block += 1
                    self.descriptions.append('')
                else:
                    if cursect.category() == 'statements':
                        cursect.extend(line)
                    elif cursect.comment:
                        comment_block += 1
                continue
            #
            # a continuation of previous item?
            if line[0].isspace() and cursect is not None and not cursect.empty():
                cursect.extend(line)
                continue
            #
            # is it a continuation of uncompleted assignment or directive?
            if cursect and not cursect.isValid():
                cursect.extend(line)
                continue
            #
            # a new line (start from first column)
            #
            # section header?
            mo = self.SECTION_HEADER.match(line)
            if mo:
                # check previous expression before a new assignment
                if cursect:
                    if not cursect.isValid():
                        parsing_errors.append(cursect.lineno, ''.join(cursect.values[:5]), 'Invalid ' + cursect.category())
                    cursect.values = []
                    cursect.finalize()
                # start a new section
                section_name = mo.group('section_name').strip()
                section_option = mo.group('section_option')
                step_names = []
                step_options = {}
                for name in section_name.split(','):
                    mo = self.SECTION_NAME.match(name)
                    if mo:
                        n, i, di, s = mo.group('name', 'index', 'default_index', 'subworkflow')
                        if n:
                            if i is None and '*' in n:
                                parsing_errors.append(lineno, line, 'Unindexed section name cannot contain wildcard character (*).')
                            step_names.append((n, i, s))
                        if di:
                            step_names.append(('default', di, s))
                    else:
                        parsing_errors.append(lineno, line, 'Invalid section name')
                if section_option is not None:
                    # this does not work directly because list parameter can have ,
                    # without having to really evaluate all complex expressions, we
                    # have to try to put syntax correctly pieces together.
                    pieces = section_option.split(',')
                    idx = 0
                    while True:
                        try:
                            # test current group
                            compile(pieces[idx].strip(), filename = '<string>', mode='exec' if '=' in pieces[idx] else 'eval')
                            # if it is ok, go next
                            idx += 1
                            if idx == len(pieces):
                                break
                        except Exception as e:
                            # error happens merge the next piece
                            if idx < len(pieces) - 1:
                                pieces[idx] += ',' + pieces[idx + 1]
                                # error happens merge the next piece
                                pieces.pop(idx + 1)
                            else:
                                # if no next group, expand previously correct one
                                if idx == 0:
                                    parsing_errors.append(lineno, line, 'Invalid section option')
                                    break
                                # break myself again
                                pieces = pieces[: idx] + pieces[idx].split(',') + pieces[idx+1:]
                                # go back
                                idx -= 1
                                pieces[idx] += '\n' + pieces[idx + 1]
                                pieces.pop(idx+1)
                    #
                    for option in pieces:
                        mo = self.SECTION_OPTION.match(option)
                        if mo:
                            opt_name, opt_value = mo.group('name', 'value')
                            #
                            #opt_value should also be a valid Python expression (with quote etc)
                            #which is most likely a string.
                            if opt_value:
                                try:
                                    opt_value = eval(opt_value)
                                except Exception as e:
                                    parsing_errors.append(lineno, line, e)
                            if opt_name == 'sigil':
                                if opt_value.count(' ') != 1 or opt_value[0] in (' ', "'") or \
                                    opt_value[-1] in (' ', "'") or \
                                    opt_value.split(' ')[0] == opt_value.split(' ')[1]:
                                    parsing_errors.append(lineno, line, 'Incorrect sigil "{}"'.format(opt_value))
                            if opt_name == 'source':
                                opt_value = [opt_value] if isinstance(opt_value, str) else opt_value
                                for sos_file in opt_value:
                                    if not os.path.isfile(sos_file) and not os.path.isfile(os.path.join(os.path.split(self.sos_script)[0], sos_file)):
                                        env.logger.warning('Source file for nested workflow {} does not exist'.format(sos_file))
                            if opt_name in step_options:
                                parsing_errors.append(lineno, line, 'Duplicate options')
                            step_options[opt_name] = opt_value
                        else:
                            parsing_errors.append(lineno, line, 'Invalid section option')
                    env.logger.trace('Header parsed with names {} and options {}'
                        .format(step_names, step_options))
                for name in step_names:
                    prev_workflows = [x[0] for x in all_step_names if '*' not in x[0]]
                    for prev_name in all_step_names:
                        # auxillary step
                        if name[1] is None and prev_name[1] is None and name[0] != prev_name[0]:
                            continue
                        # index not euqal (one of them can be None)
                        if name[1] != prev_name[1]:
                            continue
                        # index equal and one of them have wild card character
                        if '*' in name[0]:
                            names = [x for x in prev_workflows if re.match(name[0].replace('*', '.*'), x)]
                        else:
                            names = [name[0]]
                        if '*' in prev_name:
                            prev_names = [x for x in prev_workflows if re.match(prev_name[0].replace('*', '.*'), x)]
                        else:
                            prev_names = [prev_name[0]]
                        if len(set(prev_names) & set(names)):
                            parsing_errors.append(lineno, line, 'Duplicate section names')
                all_step_names.extend(step_names)
                self.sections.append(SoS_Step(step_names, step_options, is_parameters= step_names and step_names[0][0] == self._PARAMETERS_SECTION_NAME))
                cursect = self.sections[-1]
                continue
            #
            # directive?
            mo = self.DIRECTIVE.match(line)
            if mo:
                # check previous expression before a new directive
                if cursect:
                    if not cursect.isValid():
                        parsing_errors.append(cursect.lineno, ''.join(cursect.values[:5]), 'Invalid ' + cursect.category())
                    cursect.values = []
                    # allow multiple process-style actions
                    cursect.wrap_script()
                #
                directive_name = mo.group('directive_name')
                # newline should be kept in case of multi-line directive
                directive_value = mo.group('directive_value') + '\n'
                if cursect is None:
                    parsing_errors.append(lineno, line, 'Directive {} is not allowed out side of a SoS step'.format(directive_name))
                    continue
                if cursect.is_parameters:
                    parsing_errors.append(lineno, line, 'Directive {} is not allowed in {} section'.format(directive_name, self._PARAMETERS_SECTION_NAME))
                    continue
                # is it an action??
                if directive_name in self._DIRECTIVES:
                    cursect.add_directive(directive_name, directive_value, lineno)
                else:
                    # should be in string mode ...
                    cursect.add_directive('process', directive_value, lineno, action=directive_name)
                continue
            # if section is string mode?
            if cursect and cursect.isValid() and cursect.category() == 'script':
                cursect.extend(line)
                continue
            #
            # assignment?
            mo = self.ASSIGNMENT.match(line)
            if mo:
                if cursect is None:
                    self.sections.append(SoS_Step(is_global=True))
                    cursect = self.sections[-1]
                # check previous expression before a new assignment
                if not cursect.isValid():
                    parsing_errors.append(cursect.lineno, ''.join(cursect.values[:5]), 'Invalid ' + cursect.category())
                    continue
                cursect.values = []
                #
                var_name = mo.group('var_name')
                if var_name in self._DIRECTIVES:
                    parsing_errors.append(lineno, line, 'directive name cannot be used as variables')
                    continue
                # newline should be kept for multi-line assignment
                var_value = mo.group('var_value') + '\n'
                # if first line of the section, or following another assignment
                # this is assignment
                if cursect.empty() or cursect.category() == 'expression':
                    cursect.add_assignment(var_name, var_value, lineno)
                #
                # if following a directive, this can be start of an action or between directives
                elif cursect.category() == 'directive':
                    cursect.add_assignment(var_name, var_value, lineno)
                else:
                    # otherwise it is an continuation of the existing action
                    cursect.extend('{} = {}\n'.format(var_name, var_value))
                continue
            #
            # all others?
            if not cursect:
                self.sections.append(SoS_Step(is_global=True))
                cursect = self.sections[-1]
                cursect.add_statement(line, lineno)
                continue
            elif cursect.is_parameters:
                parsing_errors.append(lineno, line, 'Action statement is not allowed in {} section'.format(self._PARAMETERS_SECTION_NAME))
                continue
            #
            if cursect.empty() or cursect.category() != 'statements':
                # new statement
                cursect.add_statement(line, lineno)
            else:
                # existing one
                cursect.extend(line)

        #
        # check the last expression before a new directive
        if cursect:
            if not cursect.isValid():
                parsing_errors.append(cursect.lineno, ''.join(cursect.values[:5]), 'Invalid ' + cursect.category())
            else:
                cursect.finalize()
        #
        # if there is any parsing error, raise an exception
        if parsing_errors.errors:
            raise parsing_errors
        #
        # as the last step, let us insert the global section to all sections
        global_section = [(idx,x) for idx,x in enumerate(self.sections) if x.is_global]
        if global_section:
            global_process = ''
            for statement in global_section[0][1].statements:
                if statement[0] == '=':
                    global_process += '{} = {}\n'.format(statement[1], statement[2])
                    env.readonly_vars.add(statement[1])
                else:
                    global_process += statement[1]
            #
            for section in self.sections:
                section.global_process = global_process
            # remove the global section after inserting it to each step of the process
            self.sections.pop(global_section[0][0])


    def workflow(self, workflow_name=None, extra_sections=[]):
        '''Return a workflow with name_step+name_step specified in wf_name
        This function might be called recursively because of nested
        workflow. extra_sections are sections read from other sos script
        when the source section option is encountered. '''
        allowed_steps = None
        if not workflow_name:
            wf_name = ''
        else:
            # if consists of multiple workflows
            if '+' in workflow_name:
                wfs = []
                for wf in workflow_name.split('+'):
                    if not self.SUBWORKFLOW.match(wf):
                        raise ValueError('Incorrect workflow name {}'.format(workflow_name))
                    # if this is a combined workflow, extra_section might be specied.
                    wfs.append(self.workflow(wf, extra_sections))
                combined_wf = wfs[0]
                for wf in wfs[1:]:
                    combined_wf.extend(wf)
                return combined_wf
            # if a single workflow
            # workflow:15,-10 etc
            mo = self.SUBWORKFLOW.match(workflow_name)
            if not mo:
                raise ValueError('Incorrect workflow name {}'.format(workflow_name))
            wf_name, allowed_steps = mo.group('name', 'steps')
        # get workflow name from extra_sections
        extra_section_steps = sum([x.names for x in extra_sections if not \
            ('skip' in x.options and \
                (x.options['skip'] is None or x.options['skip'])) \
            and not ('target' in x.options)], [])
        extra_workflows = list(set([x[0] for x in extra_section_steps if '*' not in x[0]]))
        if extra_sections:
            env.logger.debug('Workflows {} imported'.format(', '.join(extra_workflows)))
        #
        if not wf_name:
            if len(self.workflows) == 1:
                wf_name = list(self.workflows)[0]
            elif 'default' in self.workflows:
                wf_name = 'default'
            else:
                raise ValueError('Name of workflow should be specified because '
                    'the script defines more than one pipelines without a default one. '
                    'Available pipelines are: {}.'.format(', '.join(self.workflows)))
        elif wf_name not in self.workflows + extra_workflows:
            raise ValueError('Workflow {} is undefined. Available workflows are: {}'.format(wf_name,
                ', '.join(self.workflows + extra_workflows)))
        # do not send extra parameters of ...
        sections = []
        # look for relevant sections in self.sections and extra sections from another script
        for section in self.sections + extra_sections:
            # skip, skip=True, skip=1 etc are all allowed.
            if 'skip' in section.options and (section.options['skip'] is None or section.options['skip']):
                continue
            if section.is_parameters:
                # include parameter only if they apply to wf_name
                if wf_name in section.names:
                    sections.append(section)
                continue
            if 'target' in section.options:
                # section global is shared by all workflows
                sections.append(section)
                continue
            for name, index, subworkflow in section.names:
                # exact match or filename like match if name contains * etc
                if fnmatch.fnmatch(wf_name, name):
                    # if there is an source option, ...
                    imported_sections = []
                    if 'source' in section.options:
                        for sos_file in section.options['source']:
                            if not os.path.isfile(sos_file):
                                sos_file = os.path.join(os.path.split(self.sos_script)[0], sos_file)
                            if not os.path.isfile(sos_file):
                                raise RuntimeError('Source file for nested workflow {} does not exist'.format(sos_file))
                            script = SoS_Script(filename=sos_file)
                            # this includes all global and parameters sections
                            imported_sections.extend(script.sections)
                    # expand subworkflow if needed
                    for i in range(len(section.names)):
                        if section.names[i][2] and isinstance(section.names[i][2], str):
                            # need to expand the workflow, but it is possible that it is nested ...
                            # so let us figure out what workflows we are expanding
                            expanded = []
                            for tmp in section.names[i][2].split('+'):
                                if ':' in tmp:
                                    tmp = tmp.split(':')[0]
                                if '_' in tmp and tmp.rsplit('_')[-1].isdigit():
                                    tmp = tmp.rsplit('_')[0]
                                expanded.append(tmp)
                            if wf_name in expanded:
                                env.logger.debug('NOT expanding {} because of potential loop'.format(section.names[i][2]))
                            else:
                                env.logger.debug('Expanding subworkflow {}'.format(section.names[i][2]))
                                # expand nested workflow
                                section.names[i] = (section.names[i][0], section.names[i][1], 
                                    self.workflow(section.names[i][2], extra_sections=extra_sections + imported_sections))
                    # now copy the step over
                    sections.append(section)
                    break
        return SoS_Workflow(wf_name, allowed_steps, sections, self.workflow_descriptions[wf_name])


    def show(self):
        textWidth = max(60, getTermWidth())
        if self.description:
            # separate \n\n
            for paragraph in dehtml(self.description).split('\n\n'):
                print('\n'.join(textwrap.wrap(paragraph, width=textWidth)))
        #
        text = 'Available workflows: {}'.format(', '.join(sorted(self.workflows)))
        print('\n' + '\n'.join(textwrap.wrap(text, width=textWidth, subsequent_indent=' '*8)))
        for workflow in sorted(self.workflows):
            wf = self.workflow(workflow)
            wf.show()
            print('')
        print('\nUse command "sos show script workflow_name" to display details of specific workflow.')

#
# subcommmand show
#
def print_traceback():
    exc_type, exc_value, exc_traceback = sys.exc_info()
    #print "*** print_tb:"
    traceback.print_tb(exc_traceback, limit=1, file=sys.stderr)
    #print "*** print_exception:"
    traceback.print_exception(exc_type, exc_value, exc_traceback,
                              limit=5, file=sys.stderr)
    #print "*** print_exc:"
    #traceback.print_exc()
    #print "*** format_exc, first and last line:"
    #formatted_lines = traceback.format_exc().splitlines()
    #print formatted_lines[0]
    #print formatted_lines[-1]
    #print "*** format_exception:"
    #print repr(traceback.format_exception(exc_type, exc_value,
    #                                      exc_traceback))
    #print "*** extract_tb:"
    #print repr(traceback.extract_tb(exc_traceback))
    #print "*** format_tb:"
    #print repr(traceback.format_tb(exc_traceback))
    #print "*** tb_lineno:", exc_traceback.tb_lineno

def sos_show(args, workflow_args):
    try:
        script = SoS_Script(filename=args.script)
        if args.workflow:
            workflow = script.workflow(args.workflow)
            workflow.show()
        else:
            script.show()
    except Exception as e:
        if args.verbosity and args.verbosity > 2:
            print_traceback()
        env.logger.error(e)
        sys.exit(1)

#
# subcommand run
#
def sos_run(args, workflow_args):
    env.verbosity = args.verbosity
    env.max_jobs = args.__max_jobs__
    # kill all remainging processes when the master process is killed.
    atexit.register(env.cleanup)
    try:
        script = SoS_Script(filename=args.script)
        workflow = script.workflow(args.workflow)
        if args.__dryrun__:
            env.run_mode = 'dryrun'
        workflow.run(workflow_args, cmd_name='{} {}'.format(args.script, args.workflow))
    except Exception as e:
        if args.verbosity and args.verbosity > 2:
            print_traceback()
        env.logger.error(e)
        sys.exit(1)

#
# subcommand dryrun
#
def sos_dryrun(args, workflow_args):
    args.__max_jobs__ = 1
    args.__dryrun__ = True
    sos_run(args, workflow_args)



#!/usr/bin/env python3
#
# This file is part of Script of Scripts (SoS), a workflow system
# for the execution of commands and scripts in different languages.
# Please visit https://github.com/bpeng2000/SOS for more information.
#
# Copyright (C) 2016 Bo Peng (bpeng@mdanderson.org)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import os
import unittest

import shlex

from pysos.utils import *

class TestUtils(unittest.TestCase):
    def testLogger(self):
        '''Test logging level'''
        for verbosity in ['0', '1', '2', '3', '4']:
            env.verbosity = verbosity
            env.logger.trace('Verbosity {}:trace message with ``empahsized text`` in between'.format(env.verbosity))
            env.logger.debug('Verbosity {}:debug message with ``empahsized text`` in between'.format(env.verbosity))
            env.logger.info('Verbosity {}:info message with ``empahsized text`` in between'.format(env.verbosity))
            env.logger.warning('Verbosity {}:warning message with ``empahsized text`` in between'.format(env.verbosity))
            env.logger.error('Verbosity {}:error message with ``empahsized text`` in between'.format(env.verbosity))
        # log
        if os.path.isfile('test.log'):
            os.remove('test.log')
        env.logfile = 'test.log'
        for verbosity in ['0', '1', '2', '3', '4']:
            env.verbosity = verbosity
            env.logger.trace('Verbosity {}:trace message with ``empahsized text`` in between'.format(env.verbosity))
            env.logger.debug('Verbosity {}:debug message with ``empahsized text`` in between'.format(env.verbosity))
            env.logger.info('Verbosity {}:info message with ``empahsized text`` in between'.format(env.verbosity))
            env.logger.warning('Verbosity {}:warning message with ``empahsized text`` in between'.format(env.verbosity))
            env.logger.error('Verbosity {}:error message with ``empahsized text`` in between'.format(env.verbosity))
        # log file should not have any color codes
        with open('test.log') as logfile:
            line_count = 0
            for line in logfile:
                line_count += 1
                self.assertFalse('\033[' in line)
            # 4 lines for all logging level (logging level of logfile is fixed to DEBUG)
            self.assertEqual(line_count, 20)

    def testInterpolation(self):
        '''Test string interpolation'''
        env.locals = {
            'a': 100,
            'b': 20,
            'c': ['file1', 'file2', 'file3'],
            'd': {'a': 'file1', 'b':'file2'},
            'e': set([1, 'a']),
            'var1': 1/2.,
            'var2': [1, 2, 3.1],
            'file': 'a/b.txt',
            'files2': ['a/b.txt', 'c.d.txt'],
            'file_ws': ['d i r/f .txt']
            }
        for sigil in ('${ }', '{ }', '[ ]', '%( )', '[[ ]]', '%( )s'):
            l, r = sigil.split(' ')
            for expr, result in [
                ('{0}1{1}', '1'),
                ('{0}a{1}', '100'),
                ('{0}a+b{1}', '120'),
                ('{0}a+b*5{1}', '200'),
                ('{0}a+b*5{1} is 200', '200 is 200'),
                ('{0}a+b*5{1} and {0}a{1}', '200 and 100'),
                ('Pre {0}a+b*5{1} and {0}a{1} after', 'Pre 200 and 100 after'),
                ('Nested {0}a+b*{0}b//2{1}{1}', 'Nested 300'),
                ('Format {0}a:.5f{1}', 'Format 100.00000'),
                ('{0}var2[:2]{1}', '1 2'),
                ('{0}var2[1:]{1}', '2 3.1'),
                # nested
                ('Nested {0}a:.{0}4+1{1}f{1}', 'Nested 100.00000'),
                # deep nested
                ('Triple Nested {0}a:.{0}4+{0}5//5{1}{1}f{1}', 'Triple Nested 100.00000'),
                # nested invalid
                ('Nested invalid {0}"{0}a-{1}"{1}', 'Nested invalid {}a-{}'.format(l, r)),
                ('Nested valid {0}"{0}a{1}-"{1}', 'Nested valid 100-'),
                #
                ('Dict {0}d{1}', ['Dict a b', 'Dict b a']),
                ('set {0}e{1}', ['set 1 a', 'set a 1']),
                ('Fmt {0}var1:.2f{1}', 'Fmt 0.50'),
                ('Fmt {0}var2:.2f{1}', 'Fmt 1.00 2.00 3.10'),
                ('LC {0}[x*2 for x in var2]{1}', 'LC 2 4 6.2'),
                ('LC {0}[x*2 for x in var2]:.2f{1}', 'LC 2.00 4.00 6.20'),
                #
                # [['a':'b', 'c':'d']['a']] works because
                #     ['a':'b', 'c':'d']a
                # is invalid so SoS does not consider ['a'] as nested expression
                #
                ('Newline {0}{{"a": "b", \n"c": "d"}}["a"]{1}', 'Newline b'),
                #
                # string literal within interpolated expression
                ('Literal {0}"{0} {1}"{1}', 'Literal ' + sigil),
                #
                ("{0}' {{}} '.format(a){1}", ' 100 '),
                #
                ("{0}os.path.basename(file){1}", 'b.txt'),
                ('{0}os.path.basename(file_ws[0]){1}', 'f .txt'),
                #
                # ! conversion
                ('{0}file!r{1}', "'a/b.txt'"),
                ('{0}file!s{1}', "a/b.txt"),
                ('''{0}"a'b"!r{1}''', '"a\'b"'),
                ('''{0}'a"b'!r{1}''', "'a\"b'"),
                #
                # !q conversion (added by SoS)
                ('{0}file_ws[0]!q{1}', "'d i r/f .txt'"),
                #
                # !, joined by ,
                ('{0}var2!r,{1}', "1,2,3.1"),
                ('{0}c!r,{1}', "'file1','file2','file3'"),
                ('{0}c!,{1}', "file1,file2,file3"),
            ]:
                #print('Interpolating "{}" with sigal "{}"'.format(expr.format(l, r).replace('\n', r'\n'), sigil))
                if isinstance(result, str):
                    self.assertEqual(interpolate(expr.format(l, r), sigil=sigil), result)
                else:
                    # for cases when the order of output is not guaranteed
                    self.assertTrue(interpolate(expr.format(l, r), sigil=sigil) in result)
        #
        # locals should be the one passed to the expression
        self.assertTrue('file_ws' in interpolate('${locals().keys()}'))
        # 5:.5.0f does not work.
        self.assertRaises(InterpolationError, interpolate, '${5:.${4/2.}f}')

    def testEval(self):
        '''Test the evaluation of SoS expression'''
        env.locals = {
            'a': 100,
            'b': 'file name',
            'c': ['file1', 'file2', 'file 3'],
            'd': {'a': 'file1', 'b':'file2'},
        }
        for expression, result in [
            ('''"This is ${a+100}"''', 'This is 200'),
            ('''"${a+100}" + "${a//100}"''', "2001"),
            ('''"${c[1]}"''', 'file2'),
            ('''"${c[1:]}"''', 'file2 file 3'),
            ('''"${d}"''', ['a b', 'b a']),
            ('''"${d}"*2''', ['a ba b', 'b ab a']),
            ('''"${d['a']}"''', 'file1'),
            ('''"${b!q}"''', "'file name'"),
            ('''"${b!r}"''', "'file name'"),
            ('''"${c!q}"''', "file1 file2 'file 3'"),
            ]:
            if isinstance(result, str):
                self.assertEqual(SoS_eval(expression), result)
            else:
                self.assertTrue(SoS_eval(expression) in result)
        #
        # interpolation will only happen in string
        self.assertRaises(SyntaxError, SoS_eval, '''${a}''')

    def testWorkflowDict(self):
        '''Test workflow dict with attribute access'''
        env.reset()
        d = WorkflowDict()
        d['a'] = 1
        self.assertEqual(d['a'], 1)
        #
        d['a'] += 1
        self.assertEqual(d['a'], 2)

    def testPatternMatch(self):
        '''Test snake match's pattern match facility'''
        res = glob_wildcards('{a}-{b}.txt', ['file-1.txt', 'file-ab.txt'])
        self.assertEqual(res['a'], ['file', 'file'])
        self.assertEqual(res['b'], ['1', 'ab'])
        res = glob_wildcards('{a}-{b}.txt', ['file--ab--cd.txt'])
        self.assertEqual(res['a'], ['file--ab-'])
        self.assertEqual(res['b'], ['cd'])
        res = glob_wildcards('{path}/{to}/{file}.txt', ['/tmp//1.txt'])
        self.assertEqual(res['path'], [])
        self.assertEqual(res['to'], [])
        self.assertEqual(res['file'], [])
        res = glob_wildcards('{path}/{to}/{file}.txt', ['/tmp/test/1.txt.txt'])
        self.assertEqual(res['path'], ['/tmp'])
        self.assertEqual(res['to'], ['test'])
        self.assertEqual(res['file'], ['1.txt'])

    def testContextStack(self):
        '''Test context stack '''
        env.locals = WorkflowDict()
        env.locals['a'] = 5
        self.assertEqual(interpolate('${a}'), '5')
        with env.push_context({'a': 10}):
            self.assertEqual(interpolate('${a}'), '10')
            with env.push_context({'a': 15}):
                self.assertEqual(interpolate('${a}'), '15')
            self.assertEqual(interpolate('${a}'), '10')
        self.assertEqual(interpolate('${a}'), '5')

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
#
# This file is part of Script of Scripts (SoS), a workflow system
# for the execution of commands and scripts in different languages.
# Please visit https://github.com/bpeng2000/SOS for more information.
#
# Copyright (C) 2016 Bo Peng (bpeng@mdanderson.org)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
import os
import sys
import re
import copy
import types
import logging
import signal
import glob
import collections
import hashlib
import shutil
import token
from tokenize import generate_tokens, untokenize
from io import StringIO
from shlex import quote
from html.parser import HTMLParser

try:
    import psutil
except ImportError as e:
    pass

#
# Logging
#

class ColoredFormatter(logging.Formatter):
    ''' A logging formatter that uses color to differntiate logging messages
    and emphasize texts. Texts that would be empahsized are quoted with
    double backslashes (`` ``).
    '''
    def __init__(self, msg):
        logging.Formatter.__init__(self, msg)
        #
        # color for different logging levels. The current terminal color
        # is used for INFO
        self.LEVEL_COLOR = {
            'TRACE': 'DARK_CYAN',
            'DEBUG': 'BLUE',
            'WARNING': 'PURPLE',
            'ERROR': 'RED',
            'CRITICAL': 'RED_BG',
        }
        self.COLOR_CODE={
            'ENDC':0,  # RESET COLOR
            'BOLD':1,
            'UNDERLINE':4,
            'BLINK':5,
            'INVERT':7,
            'CONCEALD':8,
            'STRIKE':9,
            'GREY30':90,
            'GREY40':2,
            'GREY65':37,
            'GREY70':97,
            'GREY20_BG':40,
            'GREY33_BG':100,
            'GREY80_BG':47,
            'GREY93_BG':107,
            'DARK_RED':31,
            'RED':91,
            'RED_BG':41,
            'LIGHT_RED_BG':101,
            'DARK_YELLOW':33,
            'YELLOW':93,
            'YELLOW_BG':43,
            'LIGHT_YELLOW_BG':103,
            'DARK_BLUE':34,
            'BLUE':94,
            'BLUE_BG':44,
            'LIGHT_BLUE_BG':104,
            'DARK_MAGENTA':35,
            'PURPLE':95,
            'MAGENTA_BG':45,
            'LIGHT_PURPLE_BG':105,
            'DARK_CYAN':36,
            'AUQA':96,
            'CYAN_BG':46,
            'LIGHT_AUQA_BG':106,
            'DARK_GREEN':32,
            'GREEN':92,
            'GREEN_BG':42,
            'LIGHT_GREEN_BG':102,
            'BLACK':30,
        }

    def colorstr(self, astr, color):
        return '\033[{}m{}\033[{}m'.format(color, astr,
            self.COLOR_CODE['ENDC'])

    def emphasize(self, msg, level_color=0):
        # display text within `` and `` in green
        return re.sub(r'``([^`]*)``', '\033[32m\\1\033[{}m'.format(level_color), str(msg))

    def format(self, record):
        level_name = record.levelname
        if level_name in self.LEVEL_COLOR:
            level_color = self.COLOR_CODE[self.LEVEL_COLOR[level_name]]
            record.color_levelname = self.colorstr(level_name, level_color)
            record.color_name = self.colorstr(record.name, self.COLOR_CODE['BOLD'])
            record.color_msg = self.colorstr(self.emphasize(record.msg, level_color), level_color)
        else:
            # for INFO, use default color
            record.color_levelname = record.levelname
            record.color_msg = self.emphasize(record.msg)
        return logging.Formatter.format(self, record)


def shortRepr(obj):
    '''Return a short representation of obj for clarity.'''
    if isinstance(obj, (str, int, float, bool)) or (isinstance(obj, collections.Sequence) \
        and len(obj) <= 2) or len(str(obj)) < 50:
        return repr(obj)
    elif isinstance(obj, collections.Sequence): # should be a list or tuple
        return repr(obj).split(' ')[0] + ' ...] ({} items)'.format(len(obj))
    elif isinstance(obj, dict):
        first_key = obj.keys()[0]
        return '{{{!r}:{!r}, ...}} ({} items)'.format(first_key, obj[first_key], len(obj))
    else:
        return '{}...'.format(repr(obj)[:40])

#
# SoS Workflow dictionary
#
class WorkflowDict(dict):
    """A dictionary object that
    1. Generate logging message for debugging purposes.
    2. Generate warning message if ALLCAP variables are changed.
    """
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)

    def set(self, key, value):
        '''A short cut to set value to key without triggering any logging
        or warning message.'''
        self._check_readonly(key, value)
        dict.__setitem__(self, key, value)

    def update(self, obj):
        '''Redefine update to trigger logging message'''
        for k,v in obj.items():
            self._check_readonly(k, v)
        #
        dict.update(self, obj)
        for k, v in obj.items():
            if env.verbosity > 2:
                self._log(k, v)

    def __setitem__(self, key, value):
        '''Set value to key, trigger logging and warning messages if needed'''
        if env.verbosity > 2:
            self._log(key, value)
        if env.run_mode == 'dryrun':
            self._warn(key, value)
        if key in ('_input', '_output', '_depends'):
            env.logger.warning('Variable {} is set automatically by SoS. Changing its value might cause unexpected results.'.format(key))
        self.set(key, value)

    def _check_readonly(self, key, value):
        if dict.__contains__(self, key) and key in env.readonly_vars and value != dict.__getitem__(self, key):
            raise RuntimeError('Variable {} is readonly and cannot be changed from {} to {}.'
                .format(key, dict.__getitem__(self, key), value))

    def _log(self, key, value):
        env.logger.debug('``{}`` = ``{}``'.format(key, shortRepr(value)))

    def _warn(self, key, value):
        if key.isupper() and dict.__contains__(self, key) and dict.__getitem__(self, key) != value:
            env.logger.warning('Changing readonly variable {} from {} to {}'
                .format(key, dict.__getitem__(self, key), value))
        if key.startswith('_') and key not in ('_input', '_output', '_step', '_index', '_depends'):
            env.logger.warning('{}: Variables with leading underscore is reserved for SoS temporary variables.'.format(key))

    def clone_pickleable(self):
        '''Return a copy of the existing dictionary but keep only the ones that are pickleable'''
        # FIXME: not well tested
        try:
            return {x:copy.deepcopy(y) for x,y in self.items() if not callable(y) and not isinstance(y, (types.ModuleType, WorkflowDict))}
        except:
            res = {}
            for x,y in self.items():
                try:
                    res[x] = copy.deepcopy(y)
                except:
                    pass
            return res
#
# Runtime environment
#
class RuntimeEnvironments(object):
    '''A singleton object that provides runtime environment for SoS.
    Atributes of this object include:

    logger:
        a logging object

    verbosity:
        a verbosity level object that sets the verbosity level of the logger

    logfile:
        name of logfile for the logger. default to no logfile.

    '''
    _instance = None
    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(RuntimeEnvironments, cls).__new__(cls)
        return cls._instance

    def __init__(self):
        self.reset()

    def reset(self):
        # logger
        self._logger = None
        self._verbosity = 2
        self._logfile = None
        self._set_logger()
        #
        # run mode, this mode controls how SoS actions behave
        #
        self.run_mode = 'run'
        #
        # signature mode can be
        #
        # default              (save signature, skip if signature match)
        # ignore_signature     (ignore existing signature but still saves signature)
        # assert_signature     (should 
        self.sig_mode = 'default'
        #
        # local and global dictionaries used by SoS during the
        # execution of SoS workflows
        self.locals = WorkflowDict()
        self.globals = globals()
        # variables that are defined in global and parameters sections and
        # are readonly
        self.readonly_vars = set()
        self.context_stack = []
        # 
        # maximum number of concurrent jobs
        self.max_jobs = 1
        self.running_jobs = 0

    def register_process(self, pid, msg=''):
        '''Register a process used by this SoS instance. It will also be
        used to check resource used.'''
        if not os.path.isdir('.sos'):
            os.mkdir('.sos')
        self.logger.trace('Register {} {}'.format(pid, msg))
        with open('.sos/proc_{}'.format(pid), 'w') as p:
            p.write(msg)

    def deregister_process(self, pid):
        self.logger.trace('Deregister {}'.format(pid))
        if os.path.isfile('.sos/proc_{}'.format(pid)):
            os.remove('.sos/proc_{}'.format(pid))

    def cleanup(self):
        '''Clean up all running processes'''
        for p in glob.glob('.sos/proc_*'):
            pid = int(os.path.basename(p)[5:])
            try:
                env.logger.trace('Killing {} and all its children'.format(pid))
                # psutil might not exist if SoS is not properly installed
                # but we are not acting like the end of world here
                parent = psutil.Process(pid)
                for child in parent.children(recursive=True):  # or parent.children() for recursive=False
                    child.kill()
                parent.kill()
            except Exception as e:
                env.logger.debug(e)
            os.remove(p)

    class ContextStack:
        '''A context stack and pushes existing workflow dict (env.locals)
        to a stack and make the new workflow the current dict. The 
        context will be poped as soon as the with statement ends and/or
        an exception is raised.
        '''
        def __init__(self, environ, new_dict):
            #
            self.environ = environ
            self.new_dict = new_dict

        def __enter__(self):
            # archive old items
            self.environ.context_stack.append(self.environ.locals)
            self.environ.locals = self.new_dict

        def __exit__(self, etype, value, traceback):
            self.environ.locals = self.environ.context_stack.pop()

    def push_context(self, wf_dict):
        return self.ContextStack(self, wf_dict)

    #
    # attribute logger
    #
    def _set_logger(self, unused=None):
        if not hasattr(logging, 'TRACE'):
            logging.TRACE = 5
            logging.addLevelName(logging.TRACE, "TRACE")
        # create a logger, but shutdown the previous one
        if self._logger is not None:
            self._logger.handlers = []
        self._logger = logging.getLogger()
        self._logger.setLevel(logging.DEBUG)
        # output to standard output
        cout = logging.StreamHandler()
        levels = {
            0: logging.ERROR,
            1: logging.WARNING,
            2: logging.INFO,
            3: logging.DEBUG,
            4: logging.TRACE,
            None: logging.INFO
        }
        #
        cout.setLevel(levels[self._verbosity])
        cout.setFormatter(ColoredFormatter('%(color_levelname)s: %(color_msg)s'))
        self._logger.addHandler(cout)
        self._logger.trace = lambda msg, *args: self._logger._log(logging.TRACE, msg, args)
        # output to a log file
        if self._logfile is not None:
            ch = logging.FileHandler(self._logfile, mode = 'a')
            # debug informaiton and time is always written to the log file
            ch.setLevel(logging.DEBUG)
            ch.setFormatter(logging.Formatter('%(asctime)s: %(levelname)s: %(message)s'))
            self._logger.addHandler(ch)
    #
    # atribute logger
    #
    @property
    def logger(self):
        return self._logger
    #
    # attribute verbosity
    #
    def _set_verbosity(self, v):
        if v in [0, 1, 2, 3, 4]:
            self._verbosity = v
            # reset logger to appropriate logging level
            self._set_logger()
    #
    verbosity = property(lambda self: self._verbosity, _set_verbosity)
    #
    # attribute logfile
    #
    def _set_logfile(self, f):
        self._logfile = f
        # reset logger to include log file
        self._set_logger()
    #
    logfile = property(lambda self: self._logfile, _set_logfile)

   
# set up environment variable and a default logger
env = RuntimeEnvironments()


#
# String formatting
#


from array import array
try:
    from fcntl import ioctl
    import termios
except ImportError:
    pass

def getTermWidth():
    '''Get the width of current user terminal to properly wrap SoS
    output when well-formatted output is required.
    '''
    try:
        h, w = array('h', ioctl(sys.stderr, termios.TIOCGWINSZ, '\0' * 8))[:2]
        return w
    except:
        return 78

class _DeHTMLParser(HTMLParser):
    '''This parser analyzes input text, removes HTML tags such as
    <p>, <br>, <ul>, <li> etc and returns properly formatted texts.
    '''
    def __init__(self):
        HTMLParser.__init__(self)
        self.__text = []

    def handle_data(self, data):
        text = data.strip()
        if len(text) > 0:
            text = re.sub('[ \t\r\n]+', ' ', text)
            self.__text.append(text + ' ')

    def handle_starttag(self, tag, attrs):
        if tag == 'p':
            self.__text.append('\n\n\n\n')
        elif tag == 'br':
            self.__text.append('\n\n')
        elif tag == 'ul':
            self.__text.append('')
        elif tag == 'li':
            self.__text.append('\n\n  * ')

    def handle_endtag(self, tag):
        if tag == 'ul':
            self.__text.append('\n\n')
        if tag == 'li':
            self.__text.append('\n\n')

    def handle_startendtag(self, tag, attrs):
        if tag == 'br':
            self.__text.append('\n\n')

    def text(self):
        return ''.join(self.__text).strip()

def dehtml(text):
    '''Remove HTML tag in input text and format the texts
    accordingly. '''
    try:
        parser = _DeHTMLParser()
        parser.feed(text)
        parser.close()
        return parser.text()
    except Exception as e:
        env.logger.warning('Failed to dehtml text: {}'.format(e))
        return text

#
# String interpolation and expression evluation
#

# exception classes
class Error(Exception):
    '''Base class for SoS_ScriptParser exceptions.'''

    def _get_message(self):
        '''Getter for 'message'; needed only to override deprecation in
        BaseException.'''
        return self.__message

    def _set_message(self, value):
        '''Setter for 'message'; needed only to override deprecation in
        BaseException.'''
        self.__message = value

    # BaseException.message has been deprecated since Python 2.6.  To prevent
    # DeprecationWarning from popping up over this pre-existing attribute, use
    # a new property that takes lookup precedence.
    message = property(_get_message, _set_message)

    def __init__(self, msg=''):
        self.message = msg
        Exception.__init__(self, msg)

    def __repr__(self):
        return self.message

    __str__ = __repr__


class InterpolationError(Error):
    '''Exception raised for interpolation related errors'''
    def __init__(self, text, msg):
        if '\n' in text:
            # if there is newline in original text, output it in separate
            # lines
            msg = '{}:\n{}\n'.format(msg, text)
        else:
            msg = '{}: {}'.format(msg, text)
        Error.__init__(self, msg)
        self.args = (text, msg)

#
# String intepolation
#
class SoS_String:
    '''This class implements SoS string that support interpolation using
    a local and a global dictionary. It is similar to format string of
    Python 3.6 with configurable sigil (default to ${ }) and allows the
    resulting object of types other than simple int, None etc. It also supports
    a special conversion flag !q that properly quote a filename so that
    it can be used safely within a shell script. '''

    # Format specifier that can be used at the end of the string to convert
    # result (!s|r|q) and control output format (:.2f etc). The pattern
    # is constructed according to Python format mini language.
    _FORMAT_SPECIFIER_TMPL = r'''
        ^                                   # start of expression
        (?P<expr>.*?)                       # any expression
        (?P<specifier>
        (?P<conversion>!\s*                 # conversion starting with !
        [s|r|q]?                            # conversion, q is added by SoS
        (?P<sep>,)?                         # optional character to join sequences
        )?
        (?P<format_spec>:\s*                # format_spec starting with :
        (?P<fill>.?[<>=^])?                 # optional fill|align
        (?P<sign>[-+ ])?                    # optional sign
        \#?                                 #
        0?                                  #
        (?P<width>\d+)?                     # optional width
        (?P<precision>\.\d+)?               # optional precision
        (?P<type>[bcdeEfFgGnosxX%])?        # optional type
        )?                                  # optional format_spec
        )?                                  # optional specifier
        \s*$                                # end of tring
        '''

    # DOTALL makes . matchs also to newline so this supports multi-line expression
    FORMAT_SPECIFIER = re.compile(_FORMAT_SPECIFIER_TMPL, re.VERBOSE | re.DOTALL)

    def __init__(self, sigil = '${ }'):
        # do not check sigil here because the function will be called quite frequently
        # the sigil will be checked when it is entered in SoS script.
        self.l, self.r = sigil.split(' ')
        self.error_count = 0

    def interpolate(self, text):
        '''Intepolate string with local and global dictionary'''
        #
        # We could potentially parse the text and find all interpolation text,
        # but we cannot really do it because of possible nested interpolation
        #
        # split by left sigil
        #
        # '${a} part1 ${ expr2 ${ nested }} and another ${expr2 {}} and done'
        #
        # '' 'a} part1 ' ' expr2 ' 'nested }} and another' 'expr2 {}} and done'
        #
        # 'in' test is 10 times faster than split so we do this test first.
        if self.l not in text:
            return text
        pieces = text.split(self.l, 1)
        # the first piece must be before sigil and be completed text
        #env.logger.trace('"{}" interpolated to "{}"'.format(text, res))
        return pieces[0] + self._interpolate(pieces[1])

    def _interpolate(self, text, start_nested=0):
        '''Intepolate an expression with unknown ending location. We cannot split
        the string because of potentially nested expression. start_nested indicates
        that the location of nested expression can only happen after this point.
        '''
        # no matching }, must be wrong
        if self.r not in text:
            raise InterpolationError(text[:20], "Missing {}".format(self.r))
        #
        # location of first ending sigil
        i = text.index(self.r)
        #
        # substr contains ${
        if self.l in text[start_nested:]:
            k = text.index(self.l)
        else:
            # k is very far away
            k = len(text) + 100
        #
        # nested
        if k < i:
            #                     i
            # something  ${ nested} }
            #            k
            #
            try:
                # expolate string recursively.
                return self._interpolate(text[:k] + self._interpolate(text[k + len(self.l):]))
            except Exception as e:
                self.error_count += 1
                if self.error_count > 10:
                    raise
                # This is for the case where inner sigil is actually part of the syntax. For example, if
                # sigil = []
                #
                #     [[x*2 for x in [a, b]]]
                #
                # will first try to evaluate
                #
                #     x*2 for x in [a, b]
                #
                # without success. This part will then try to evaluate
                #
                #     [x*2 x for x in [a, b]]
                #
                # namely keeping [] as python expression
                #
                return self._interpolate(text, start_nested=k + len(self.l))
        else:
            # non-nested case, proceed from left to right
            #
            #            i
            # something {} } ${ another }
            #                k
            j = i
            while True:
                try:
                    # is syntax correct?
                    mo = self.FORMAT_SPECIFIER.match(text[:j])
                    if mo:
                        expr, fmt, sep = mo.group('expr', 'specifier', 'sep')
                        if sep:
                            if fmt.startswith('!,'):
                                fmt = fmt[2:]
                            else:
                                fmt = fmt[:2] + fmt[3:]
                        else:
                            sep = ' '
                    else:
                        expr = text[:j]
                        fmt = None
                        sep = None
                    # if the syntax is correct
                    compile(expr, '<string>', 'eval')
                    try:
                        result = eval(expr, env.globals, env.locals)
                    except Exception as e:
                        raise InterpolationError(expr, e)
                    # evaluate the expression and interpolate the next expression
                    return self._repr(result, fmt, sep) + self.interpolate(text[j+len(self.r):])
                except Exception as e:
                    self.error_count += 1
                    if self.r not in text[j+1:]:
                        raise InterpolationError(text[:j], e)
                    j = text.index(self.r, j+1)
                    #                           j
                    # something {} } ${ another }
                    #                k
                    if j > k:
                        return self._interpolate(text[:k] + self._interpolate(text[k+len(self.l):]))

    def _format(self, obj, fmt):
        '''Format an object in basic type (not list etc)
        '''
        # handling special !q conversion flag
        if fmt.startswith('!q'):
            # special SoS conversion for shell quotation.
            return self._format(quote(obj), fmt[2:])
        else:
            # use
            return ('{' + fmt + '}').format(obj)

    def _repr(self, obj, fmt=None, sep=' '):
        '''Format an object. fmt will be applied to all elements if obj is not
        in a basic type. Callable object cannot be outputed (an InterpolationError
        will be raised).
        '''
        if isinstance(obj, str):
            return obj if fmt is None else self._format(obj, fmt)
        elif isinstance(obj, collections.Iterable):
            # the object might be nested...
            return sep.join([self._repr(x, fmt, sep) for x in obj])
        elif isinstance(obj, collections.Callable):
            raise InterpolationError(repr(obj), 'Cannot interpolate callable object.')
        else:
            return repr(obj) if fmt is None else self._format(obj, fmt)

def interpolate(text, sigil='${ }'):
    '''Evaluate expressions in `text` marked by specified `sigil` using provided
    global and local dictionaries, and replace the expressions with their formatted strings.'''
    return SoS_String(sigil).interpolate(text)


def ConvertString(s, sigil):
    '''Convert a unicode string to a raw string and interpolate expressions
    within it by parsing the python expression and statement BEFORE they are
    evaluated (or executed).

    FIXME: the expression might have a dynamic option which should prevent
    string interpolation. Not sure how to handle this option right now.
    '''
    result = []
    left_sigil = sigil.split(' ')[0]
    # tokenize the input syntax.
    if sys.version_info.major == 2:
        g = generate_tokens(StringIO(s.decode()).readline)
    else:
        # python 3 string is already unicode string
        g = generate_tokens(StringIO(s).readline)
    for toknum, tokval, _, _, _  in g:
        if toknum == token.STRING:
            # if this item is a string that uses triple single quote
            if tokval.startswith("'''"):
                # we convert it to a raw string
                tokval = u'r' + tokval
            # we then perform interpolation on the string and put it back to expression
            if left_sigil in tokval:
                tokval = 'interpolate(' + tokval + ", \'" + sigil + "')"
        # the resusting string is put back to the expression (or statement)
        result.append((toknum, tokval))
    return untokenize(result)

def SoS_eval(expr, sigil='${ }'):
    '''Evaluate an expression after modifying (convert ' ' string to raw string,
    interpolate expressions) strings.'''
    expr = ConvertString(expr, sigil)
    return eval(expr, env.globals, env.locals)

def SoS_exec(stmts, sigil='${ }'):
    '''Execute a statement after modifying (convert ' ' string to raw string,
    interpolate expressions) strings.'''
    # the trouble here is that we have to execute the statements line by line
    # because the variables defined. The trouble is in cases such as class
    # definition
    #
    # class A:
    #     def __init__(self):
    #         pass
    # # this is already correct syntax but the remaining piece is not.
    #     def another_one(self):
    #         pass
    #
    # We therefore has to be a bit more clever on this.
    #
    # we first group all lines by their own group
    # we then try to find syntaxly valid groups
    code_group = [x for x in stmts.split('\n')]
    idx = 0
    while True:
        try:
            # test current group
            compile(code_group[idx], filename = '<string>', mode='exec')
            # if it is ok, go next
            idx += 1
            if idx == len(code_group):
                break
        except:
            # error happens merge the next line
            if idx < len(code_group) - 1:
                code_group[idx] += '\n' + code_group[idx + 1]
                code_group.pop(idx + 1)
            else:
                # if no next group, expand previously correct one
                if idx == 0:
                    raise RuntimeError('Failed to find syntax correct group: {}'.format(stmts))
                # break myself again
                code_group = code_group[: idx] + code_group[idx].split('\n') + code_group[idx+1:]
                # go back
                idx -= 1
                code_group[idx] += '\n' + code_group[idx + 1]
                code_group.pop(idx+1)
    #
    # execute statements one by one
    executed = ''
    for code in code_group:
        stmts = ConvertString(code, sigil)
        exec(stmts, env.globals, env.locals)
        executed += stmts + '\n'
    #
    env.logger.trace('Executed\n{}'.format(executed))
    return executed



#
# Runtime signature
#

def textMD5(text):
    '''Get md5 of a piece of text'''
    m = hashlib.md5()
    m.update(text.encode())
    return m.hexdigest()

def partialMD5(filename):
    '''Calculate partial MD5, basically the first and last 32M
    of the file for large files. This should signicicantly reduce
    the time spent on the creation and comparison of file signature
    when dealing with large bioinformat ics datasets. '''
    filesize = os.path.getsize(filename)
    # calculate md5 for specified file
    md5 = hashlib.md5()
    block_size = 2**20  # buffer of 1M
    try:
        if filesize < 2**26:
            with open(filename, 'rb') as f:
                while True:
                    data = f.read(block_size)
                    if not data:
                        break
                    md5.update(data)
        else:
            count = 64
            # otherwise, use the first and last 32M
            with open(filename, 'rb') as f:
                while True:
                    data = f.read(block_size)
                    count -= 1
                    if count == 32:
                        f.seek(-2**25, 2)
                    if not data or count == 0:
                        break
                    md5.update(data)
    except IOError as e:
        sys.exit('Failed to read {}: {}'.format(filename, e))
    return md5.hexdigest()


class RuntimeInfo:
    '''Record run time information related to a number of output files. Right now only the
    .exe_info files are used.
    '''
    def __init__(self, script, input_files=[], output_files=[], dependent_files = [], pid=None, workdir='.'):
        '''Runtime information for specified output files
        output_files:
            intended output file

        pid:
            process id.

        workdir:
            Current working directory.,
        '''
        self.script = script if isinstance(script, str) else ''.join(script)
        self.input_files = [input_files] if isinstance(input_files, str) else input_files
        self.output_files = [output_files] if isinstance(output_files, str) else output_files
        self.dependent_files = dependent_files if isinstance(dependent_files, str) else dependent_files
        #
        sig_name = os.path.realpath(os.path.expanduser(self.output_files[0])) + textMD5('{} {} {} {}'.format(script, input_files, output_files, dependent_files))
        #
        if not output_files:
            self.sig_file = None
            self.proc_out = None
            self.proc_err = None
            self.proc_lck = None
            self.proc_info = None
            self.proc_cmd = None
            self.proc_prog = None
            self.proc_done = None
            self.manifest = None
        else:
            #
            # If the output path is outside of the current working directory
            rel_path = os.path.relpath(sig_name, os.path.realpath(workdir))
            # if this file is not relative to cache, use global signature file
            if rel_path.startswith('../'):
                self.sig_file = os.path.join(os.path.expanduser('~/.sos/.runtime'), sig_name.lstrip(os.sep))
            else:
                # if this file is relative to cache, use local directory
                self.sig_file = os.path.join('.sos/.runtime', rel_path)
            # path to file
            sig_path = os.path.split(self.sig_file)[0]
            if not os.path.isdir(sig_path):
                try:
                    os.makedirs(sig_path)
                except Exception as e:
                    raise RuntimeError('Failed to create runtime directory {}: {}'.format(sig_path, e))
            env.logger.trace('Using signature file {} for output {}'.format(self.sig_file, output_files))
            if pid is None:
                self.pid = os.getpid()
            else:
                self.pid = pid
            self.proc_out = '{}.out_{}'.format(self.sig_file, self.pid)
            self.proc_err = '{}.err_{}'.format(self.sig_file, self.pid)
            self.proc_lck = '{}.lck'.format(self.sig_file)
            self.proc_info = '{}.exe_info'.format(self.sig_file)
            self.proc_cmd = '{}.cmd'.format(self.sig_file)
            self.proc_done = '{}.done_{}'.format(self.sig_file, self.pid)
            self.proc_prog = '{}.working_{}'.format(self.sig_file, self.pid)
            self.manifest = '{}.manifest'.format(self.sig_file)

    def write(self):
        '''Write .exe_info file with signature of script, input, output and dependent files.'''
        if not self.proc_info:
            return
        env.logger.trace('Write signature {}'.format(self.proc_info))
        with open(self.proc_info, 'w') as md5:
            md5.write('{}\n'.format(textMD5(self.script)))
            for f in self.input_files + self.output_files + self.dependent_files:
                f = os.path.realpath(os.path.expanduser(f))
                md5.write('{}\t{}\n'.format(f, partialMD5(f)))

    def validate(self):
        '''Check if ofiles and ifiles match signatures recorded in md5file'''
        if not self.proc_info or not os.path.isfile(self.proc_info):
            env.logger.trace('Fail because of no signature file {}'.format(self.proc_info))
            return False
        #
        # duplicated files are only tested once.
        all_files = [os.path.realpath(os.path.expanduser(x)) for x in set(self.input_files + self.output_files + self.dependent_files)]
        # file not exist?
        if not all(os.path.isfile(x) for x in all_files):
            env.logger.trace('Fail because of missing one of the files {}'.format(', '.join(all_files)))
            return False
        #
        files_checked = {x:False for x in all_files}
        with open(self.proc_info) as md5:
            cmdMD5 = md5.readline().strip()   # command
            if textMD5(self.script) != cmdMD5:
                env.logger.trace('Fail because of command change')
                return False
            for line in md5:
                try:
                    f, m = line.rsplit('\t', 1)
                    f = os.path.realpath(os.path.expanduser(f))
                except Exception as e:
                    env.logger.error('Wrong md5 line {} in {}: {}'.format(line, md5file, e))
                    continue
                if f not in files_checked:
                    env.logger.waring('{} not need to be checked'.format(f))
                    continue
                if partialMD5(f) != m.strip():
                    env.logger.error('MD5 mismatch {}'.format(f))
                    return False
                files_checked[f] = True
        #
        if not all(files_checked.values()):
            env.logger.warning('No MD5 signature for {}'.format(', '.join(x for x,y in files_checked.items() if not y)))
            return False
        return True

    def clear(self, types=['out', 'err', 'done']):
        if self.sig_file is None:
            return
        for filename in sum([glob.glob(self.sig_file + '.{}_*'.format(x)) for x in types], []):
            try:
                os.remove(filename)
            except Exception as e:
                env.logger.warning('Fail to remove {}: {}'.format(filename, e))


#
# The following is adapted from snakemake so that the syntax of the pattern parameter
# would be the same as snakemake.
#
# https://bitbucket.org/snakemake/snakemake/src/22eff35627401d3bb243e068c5dce97107e7090b/snakemake/io.py?at=master&fileviewer=file-view-default
#
# I will re-implement it if there is any license issue with the code
#

_wildcard_regex = re.compile(
    "\{\s*(?P<name>\w+?)(\s*,\s*(?P<constraint>([^\{\}]+|\{\d+(,\d+)?\})*))?\s*\}")


def regex(filepattern):
    f = []
    last = 0
    wildcards = set()
    for match in _wildcard_regex.finditer(filepattern):
        f.append(re.escape(filepattern[last:match.start()]))
        wildcard = match.group("name")
        if wildcard in wildcards:
            if match.group("constraint"):
                raise ValueError(
                    "If multiple wildcards of the same name "
                    "appear in a string, eventual constraints have to be defined "
                    "at the first occurence and will be inherited by the others.")
            f.append("(?P={})".format(wildcard))
        else:
            wildcards.add(wildcard)
            f.append("(?P<{}>{})".format(wildcard, match.group("constraint") if
                                         match.group("constraint") else ".+"))
        last = match.end()
    f.append(re.escape(filepattern[last:]))
    f.append("$")  # ensure that the match spans the whole file
    return "".join(f)


def glob_wildcards(pattern, files=None):
    """
    Glob the values of the wildcards by matching the given pattern to the filesystem.
    Returns a named tuple with a list of values for each wildcard.
    """
    pattern = os.path.normpath(pattern)
    first_wildcard = re.search("{[^{]", pattern)
    dirname = os.path.dirname(pattern[:first_wildcard.start(
    )]) if first_wildcard else os.path.dirname(pattern)
    if not dirname:
        dirname = "."

    names = [match.group('name')
             for match in _wildcard_regex.finditer(pattern)]
    res = {x: [] for x in names}
    pattern = re.compile(regex(pattern))

    if files is None:
        files = ((os.path.join(dirpath, f) if dirpath != "." else f)
                 for dirpath, dirnames, filenames in os.walk(dirname)
                 for f in chain(filenames, dirnames))

    for f in files:
        match = re.match(pattern, f)
        if match:
            for name, value in match.groupdict().items():
                res[name].append(value)
    return res

def apply_wildcards(pattern,
                    wildcards,
                    fill_missing=False,
                    fail_dynamic=False,
                    dynamic_fill=None,
                    keep_dynamic=False):
    def format_match(match):
        name = match.group("name")
        try:
            value = wildcards[name]
            if fail_dynamic and value == dynamic_fill:
                raise WildcardError(name)
            return str(value)  # convert anything into a str
        except KeyError as ex:
            if keep_dynamic:
                return "{{{}}}".format(name)
            elif fill_missing:
                return dynamic_fill
            else:
                raise RuntimeError('Wildcard apply error {}'.format(ex))

    return re.sub(_wildcard_regex, format_match, pattern)




#!/usr/bin/env python3
#
# This file is part of Script of Scripts (SoS), a workflow system
# for the execution of commands and scripts in different languages.
# Please visit https://github.com/bpeng2000/SOS for more information.
#
# Copyright (C) 2016 Bo Peng (bpeng@mdanderson.org)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

# passing string as unicode to python 2 version of SoS
# to ensure compatibility
from __future__ import unicode_literals

import os
import unittest

from pysos import *

class TestParser(unittest.TestCase):
    def setUp(self):
        if hasattr(env, 'locals'):
            env.locals.clear()
        env.protect_vars_assigned = False

    def testFileFormat(self):
        '''Test recognizing the format of SoS script'''
        # file format must be 'fileformat=SOSx.x'
        self.assertRaises(ParsingError, SoS_Script,
            '#fileformat=SS2')
        self.assertRaises(ParsingError, SoS_Script,
            '#fileformat=SOS1.0beta')
        #
        # parse a larger script with gormat 1.1
        script = SoS_Script(filename='scripts/section1.sos')
        # not the default value of 1.0
        self.assertEqual(script.format_version, '1.1')

    def testStringLiteral(self):
        '''Test string literals of SoS'''
        script = SoS_Script(r"""
[0]
a = 'a\n'
b = "a\n"
c = '''a\n

b'''
""")
        wf = script.workflow()
        wf.run()
        self.assertEqual(env.locals['a'], 'a\n')
        self.assertEqual(env.locals['b'], 'a\n')
        # MAYJOR difference
        self.assertEqual(env.locals['c'], 'a\\n\nb')
        script = SoS_Script(r'''
[0]
c = """a\n

b"""
''')
        wf = script.workflow()
        wf.run()
        # Note the difference between """ and ''' quotes
        self.assertEqual(env.locals['c'], 'a\n\nb')

    def testWorkflows(self):
        '''Test workflows defined in SoS script'''
        script = SoS_Script('''[0]''')
        self.assertEqual(sorted(script.workflows), ['default'])
        script = SoS_Script('''[0]\n[1]''')
        self.assertEqual(sorted(script.workflows), ['default'])
        script = SoS_Script('''[0]\n[*_1]''')
        self.assertEqual(sorted(script.workflows), ['default'])
        script = SoS_Script('''[0]\n[*_1]\n[auxillary:target='*.txt']''')
        self.assertEqual(sorted(script.workflows), ['default'])
        script = SoS_Script('''[0]\n[*_1]\n[human_1]''')
        self.assertEqual(sorted(script.workflows), ['default', 'human'])
        script = SoS_Script('''[0]\n[*_1]\n[human_1]\n[mouse_2]''')
        self.assertEqual(sorted(script.workflows), ['default', 'human', 'mouse'])
        script = SoS_Script('''[0]\n[*_1]\n[human_1]\n[mouse_2]\n[s*_2]''')
        self.assertEqual(sorted(script.workflows), ['default', 'human', 'mouse'])
        # skip option
        script = SoS_Script('''[0]\n[*_1]\n[human_1]\n[mouse_2:skip]\n[s*_2]''')
        self.assertEqual(sorted(script.workflows), ['default', 'human'])
        # unnamed
        script = SoS_Script('''[0]\n[*_1]\n[human_1]\n[mouse]\n[s*_2]''')
        self.assertEqual(sorted(script.workflows), ['default', 'human', 'mouse'])

    def testSections(self):
        '''Test section definitions'''
        # bad names
        for badname in ['56_1', '_a', 'a_', '1x', '*', '?']:
            self.assertRaises(ParsingError, SoS_Script, '[{}]'.format(badname))
        # bad options
        for badoption in ['ss', 'skip a', 'skip:_', 'skip, skip']:
            self.assertRaises(ParsingError, SoS_Script, '[0:{}]'.format(badoption))
        # option value should be a valid python expression
        for badoption in ['sigil=a', 'alias=a', 'sigil="[]"', 'sigil="| |"']:
            self.assertRaises(ParsingError, SoS_Script, '[0:{}]'.format(badoption))
        # good options
        for goodoption in ['sigil="[ ]"', 'alias="a"']:
            SoS_Script('[0:{}]'.format(goodoption))
        # allowed names
        for name in ['a5', 'a_5', '*_0', 'a*1_100']:
            SoS_Script('[{}]'.format(name))
        # no directive in global section
        self.assertRaises(ParsingError, SoS_Script,
            '''input: 'filename' ''')
        # duplicate sections
        self.assertRaises(ParsingError, SoS_Script,
            '''[1]\n[1]''')
        self.assertRaises(ParsingError, SoS_Script,
            '''[1]\n[3]\n[2,1]''')
        self.assertRaises(ParsingError, SoS_Script,
            '''[a_1]\n[a_3]\n[*_1]''')
        # no duplicated section header
        SoS_Script('''[a_1]\n[a_3]\n[b*_1]''')

    def testGlobalVariables(self):
        '''Test definition of variables'''
        # global section cannot have directive
        self.assertRaises(ParsingError, SoS_Script,
            '''input: 'filename' ''')
        # or unrecognized directive
        self.assertRaises(ParsingError, SoS_Script,
            '''inputs: 'filename' ''')
        # allow definition
        SoS_Script('''a = '1' ''')
        SoS_Script('''a = ['a', 'b'] ''')
        # but this one has incorrect syntax
        self.assertRaises(ParsingError, SoS_Script,
            '''a = 'b  ''')
        # This one also does not work because b is not defined.
        delattr(env, 'locals')
        script = SoS_Script('''a = b\n[0] ''')
        wf = script.workflow()
        self.assertRaises(RuntimeError, wf.run)
        # multi-line string literal
        SoS_Script('''a = """
this is a multi line
string """
''')
        # multi-line list literal, even with newline in between
        SoS_Script('''a = [
'a',

'b'
]
''')
        #
        script = SoS_Script(filename='scripts/section1.sos')
        # not the default value of 1.0

    def testParameters(self):
        '''Test parameters section'''
        # directive not allowed in parameters
        self.assertRaises(ParsingError, SoS_Script,
            '''
[parameters]
input: 'filename' 
''')
        self.assertRaises(ParsingError, SoS_Script,
            '''
[parameters]
func()
''')    
        script = SoS_Script(filename='scripts/section1.sos')
        self.assertRaises(ArgumentError, script.workflow('chapter_0').run,
            args=['--not_exist'])
        self.assertRaises(ArgumentError, script.workflow('chapter_0').run,
            args=['--par1', 'a', 'b'])
        # 
        # test parameter using global definition
        script = SoS_Script('''
a="100"

[parameters]
b=str(int(a)+1)
''')
        script.workflow().run()
        self.assertEqual(env.locals['b'], '101')
        env.locals.clear()
        script = SoS_Script('''
a=100

[parameters]
b=a+1
''')
        script.workflow().run()
        self.assertEqual(env.locals['b'], 101)
        #
        script = SoS_Script('''
a=100

[parameters]
b=a+1
''')
        wf = script.workflow()
        self.assertRaises(ArgumentError, wf.run, args=['--b', 'a'])
        #
        script = SoS_Script('''
a=100

[parameters]
b=a+1.
''')
        wf = script.workflow()
        wf.run(args=['--b', '1000'])
        #
        self.assertEqual(env.locals['b'], 1000)
        #
        # test string interpolation of the parameter section
        script = SoS_Script('''
a=100

[parameters]
b='${a+1}'
''')
        wf = script.workflow()
        wf.run()
        self.assertEqual(env.locals['b'], '101')
        # test alternative sigil
        script = SoS_Script('''
a=100

[parameters: sigil='[ ]']
b='[a+1]'
''')
        wf = script.workflow()
        wf.run()
        self.assertEqual(env.locals['b'], '101')
        #
        # argument has hve a value
        self.assertRaises(ParsingError, SoS_Script, '''
[parameters]
b=
''')
        # if it is a type, must provide value
        script = SoS_Script('''
[parameters]
b = int
''')
        self.assertRaises(ArgumentError, script.workflow().run)
        #
        script = SoS_Script('''
[parameters]
b = list
''')
        self.assertRaises(ArgumentError, script.workflow().run)
        # also require the type
        script = SoS_Script('''
[parameters]
b = int
''')
        self.assertRaises(ArgumentError, script.workflow().run, args=['--b', 'file'])
        #
        script = SoS_Script('''
[parameters]
b = int
''')
        wf = script.workflow()
        wf.run(args=['--b', '5'])
        self.assertEqual(env.locals['b'], 5)
        # list is ok
        script = SoS_Script('''
[parameters]
b = list
''')
        wf = script.workflow()
        wf.run(args=['--b', '5'])
        self.assertEqual(env.locals['b'], ['5'])
        # bool
        script = SoS_Script('''
[parameters]
b = bool
''')
        wf = script.workflow()
        wf.run(args=['--b', 't'])
        self.assertEqual(env.locals['b'], True)
        script = SoS_Script('''
[parameters]
b = True
''')
        wf = script.workflow()
        wf.run(args=['--b', 'False'])
        self.assertEqual(env.locals['b'], False)
        script = SoS_Script('''
[parameters]
b = True
''')
        wf = script.workflow()
        wf.run(args=['--b', '1'])
        self.assertEqual(env.locals['b'], True)
        script = SoS_Script('''
[parameters]
b = bool
''')
        wf = script.workflow()
        wf.run(args=['--b', 'no'])
        self.assertEqual(env.locals['b'], False)
        #
        # should fail for undefined variables
        script = SoS_Script('''
[parameters]
a = 5
''')
        self.assertRaises(ArgumentError, script.workflow().run,
            args=['--b', 'file'])
        # and for cases without parameter section
        script = SoS_Script('''
''')        
        self.assertRaises(ArgumentError, script.workflow().run,
            args=['--b', 'file'])

    def testSectionVariables(self):
        '''Test variables in sections'''
        # directive name cannot be used as variable
        self.assertRaises(ParsingError, SoS_Script,
            '''[0]
input='a.txt' ''')
        self.assertRaises(ParsingError, SoS_Script,
            '''[0]
output='a.txt' ''')
        self.assertRaises(ParsingError, SoS_Script,
            '''[0]
depends='a.txt' ''')

    def testSectionDirectives(self):
        '''Test directives of sections'''
        # cannot be in the global section
        self.assertRaises(ParsingError, SoS_Script,
            '''input: 'filename' ''')
        # multi-line OK
        SoS_Script('''
[0]
input: 'filename',
    'filename1'

''')
        # An abusive case with multi-line OK, from first column ok, blank line ok
        SoS_Script('''
[0]
input: 'filename',
'filename1',

filename4,
opt1=value
output: 
    blah

depends:
'something else'
''')
        # option with expression ok
        SoS_Script('''
[0]
input: 'filename',  'filename2', opt=value==1

''')
        # unrecognized directive, allowed now
        SoS_Script('''
[0]
something: 'filename',  filename2, opt=value==1
''')
        # need commma
        self.assertRaises(ParsingError, SoS_Script, '''
[0]
input: 'filename'  filename2
''')
        # can be after action
        SoS_Script('''
[0]
func()        
input: 'filename',  'filename2', opt=value==1
''')
        # assignments between directives are allowed
        SoS_Script('''
[0]
input: 'filename',  'filename2', opt=value==1
a = 'some text'
output: 'filename',  'filename2', opt=value==1
''')
        # can have action between directives
        SoS_Script('''
[0]
input: 'filename',  'filename2', opt=value==1
abc
output: 'filename',  'filename2', opt=value==1
''')

    def testScriptFormat(self):
        '''Test putting scripts directly under action'''
        script = SoS_Script('''
[0]
input: 'filename',  'filename2', opt=value==1
R:

open('something')
save.put()

''')
        script = SoS_Script('''
[0]
input: 'filename',  'filename2', opt=value==1
R: concurrent = True

open('something')
save.put()

''')
        script = SoS_Script('''
[0]
input: 'filename',  'filename2', opt=value==1
R: concurrent = True,
    workdir = 'someelse else'

open('something')
save.put()

''')
        # test dedent
        script = SoS_Script('''
[0]
python:
    from pysos import env
    env.logger.warning('I am from a dented text')
    if 1:
        env.logger.warning('Another warning')
''')
        wf = script.workflow()
        wf.run()



    def testInput(self):
        '''Test input directive'''
        script = SoS_Script('''
[0]
files = ['a.txt', 'b.txt']

input: 'a.pdf', files, skip=False

''')
        env.run_mode = 'dryrun'
        script.workflow('default').run()

    def testGroupBy(self):
        '''Test group_by parameter of step input'''
        # group_by = 'all'
        script = SoS_Script('''
[0]

executed = []
input: ['a{}.txt'.format(x) for x in range(1, 5)], group_by='all'

executed.append(_input)

''')
        env.run_mode = 'dryrun'
        wf = script.workflow()
        wf.run()
        self.assertEqual(env.locals['executed'],  [['a1.txt', 'a2.txt', 'a3.txt', 'a4.txt']])
        # group_by = 'single'
        script = SoS_Script('''
[0]

executed = []
input: ['a{}.txt'.format(x) for x in range(1, 5)], group_by='single'

executed.append(_input)

''')
        env.run_mode = 'dryrun'
        wf = script.workflow()
        wf.run()
        self.assertEqual(env.locals['executed'],  [['a1.txt'], ['a2.txt'], ['a3.txt'], ['a4.txt']])
        # group_by = 'pairs'
        script = SoS_Script('''
[0]

executed = []
input: ['a{}.txt'.format(x) for x in range(1, 5)], group_by='pairs'

executed.append(_input)

''')
        env.run_mode = 'dryrun'
        wf = script.workflow()
        wf.run()
        self.assertEqual(env.locals['executed'],  [['a1.txt', 'a3.txt'], ['a2.txt', 'a4.txt']])
        # group_by = 'pairwise'
        script = SoS_Script('''
[0]

executed = []
input: ['a{}.txt'.format(x) for x in range(1, 5)], group_by='pairwise'

executed.append(_input)

''')
        env.run_mode = 'dryrun'
        wf = script.workflow()
        wf.run()
        self.assertEqual(env.locals['executed'],  [['a1.txt', 'a2.txt'], ['a2.txt', 'a3.txt'], ['a3.txt', 'a4.txt']])
        # group_by = 'combinations'
        script = SoS_Script('''
[0]

executed = []
input: ['a{}.txt'.format(x) for x in range(1, 5)], group_by='combinations'

executed.append(_input)

''')
        env.run_mode = 'dryrun'
        wf = script.workflow()
        wf.run()
        self.assertEqual(env.locals['executed'],  [['a1.txt', 'a2.txt'], ['a1.txt', 'a3.txt'], 
            ['a1.txt', 'a4.txt'], ['a2.txt', 'a3.txt'], ['a2.txt', 'a4.txt'], ['a3.txt', 'a4.txt']])


    def testSectionActions(self):
        '''Test actions of sections'''
        SoS_Script(
            """
[0]
func('''
multiline 
string''', with_option=1
)
""")
        self.assertRaises(ParsingError, SoS_Script,
            '''
[0]
func(
''')

    def testDescriptions(self):
        '''Test script and workflow descriptions'''
        script = SoS_Script('''# first block

# global
# description

# human
# description of human

# description of human continued

[human_1]

a = '1'

# mouse
# mouse description
#

[mouse_1]
''')
        self.assertEqual(script.description, 'global\ndescription\n\n')
        self.assertEqual(script.workflow('human').description, 'description of human\ndescription of human continued\n')
        self.assertEqual(script.workflow('mouse').description, 'mouse description\n')

    def testLongerCode(self):
        '''Test definition of classes (with intermediate newlines) in step.'''
        script = SoS_Script('''# first block

[0]
class A:
    def __init__(self):
        pass

    # the newline above should be fine because SoS treat this as
    # regular lines
    def __call__(self):
        return 0

b = A()()

''')
        wf = script.workflow()
        wf.run()
        self.assertEqual(env.locals['b'], 0)

    def testCombinedWorkflow(self):
        '''Test the creation and execution of combined workfow'''
        script = SoS_Script('''
a0 = 0
if 'executed' in locals():
    executed.append(_step.name)
else:
    executed = [_step.name]
[parameters]
a = a0 + 1
[a_1]
[a_2]
[a_3]
[a_4]
output: 'out_a_4'
[b_1]
input_b1 = _step.input
[b_2]
[b_3]
[b_4]
[c]
[d]
''')
        env.run_mode = 'dryrun'
        wf = script.workflow('a+b')
        wf.run()
        self.assertEqual(env.locals['executed'], ['a_parameters', 'a_1', 'a_2', 'a_3', 'a_4', 'b_parameters', 'b_1', 'b_2', 'b_3', 'b_4'])
        self.assertEqual(env.locals['a'], 1)
        self.assertEqual(env.locals['input_b1'], ['out_a_4'])
        #
        wf = script.workflow('a_ 1-2 + a_4 + b_3-')
        wf.run()
        self.assertEqual(env.locals['executed'], ['a_parameters', 'a_1', 'a_2', 'a_parameters', 'a_4', 
            'b_parameters', 'b_3', 'b_4'])
        #
        wf = script.workflow('a+c+d')
        wf.run()
        self.assertEqual(env.locals['executed'], ['a_parameters', 'a_1', 'a_2', 'a_3', 'a_4', 'c_parameters', 'c_0', 'd_parameters', 'd_0'])

    def testNestedWorkflow(self):
        '''Test the creation and execution of combined workfow'''
        script = SoS_Script('''
if 'executed' in locals():
    executed.append(_step.name)
else:
    executed = [_step.name]
if 'inputs' not in locals():
    inputs = []

[paramters]
[a_1]
inputs.append(_step.input)
[a_2]
inputs.append(_step.input)
[a_3]
inputs.append(_step.input)
[a_4]
output: 'a.done'
inputs.append(_step.input)
[b_1]
input: 'b.begin'
inputs.append(_step.input)
[b_2]
inputs.append(_step.input)
[b_3]
inputs.append(_step.input)
[b_4]
output: 'b.txt'
inputs.append(_step.input)
[c=a+b]
input: 'a.txt'
output: 'b.txt'
inputs.append(_step.input)
''')
        env.run_mode = 'dryrun'
        wf = script.workflow('c')
        wf.run()
        self.assertEqual(env.locals['executed'], ['c_0', 'a_1', 'a_2', 'a_3', 'a_4',
            'b_1', 'b_2', 'b_3', 'b_4'])
        self.assertEqual(env.locals['inputs'], [['a.txt'], ['a.txt'], [], [], [], ['b.begin'], [], [], []])
        # step will be looped
        script = SoS_Script('''
if 'executed' in locals():
    executed.append(_step.name)
else:
    executed = [_step.name]
if 'inputs' not in locals():
    inputs = []

[a_1]
output: _input[0] + '.a1'
inputs.append(_step.input)
[a_2]
output: _input[0] + '.a2'
inputs.append(_step.input)
[c=a]
input: 'a.txt', 'b.txt', group_by='single'
inputs.append(_input)
''')
        env.run_mode = 'dryrun'
        wf = script.workflow('c')
        wf.run()
        self.assertEqual(env.locals['executed'], ['c_0', 'a_1', 'a_2', 'a_1', 'a_2'])
        #self.assertEqual(env.locals['inputs'], [['a.txt'], ['a.txt'], ['a.txt.a1'], ['b.txt'], ['b.txt'], ['b.txt.a1']])
        #
        # allow specifying a single step
        # step will be looped
        script = SoS_Script('''
if 'executed' in locals():
    executed.append(_step.name)
else:
    executed = [_step.name]
[a_1]
[a_2]
[c_0]
[c_1=a_2]
input: 'a.txt', 'b.txt', group_by='single'
''')
        env.run_mode = 'dryrun'
        wf = script.workflow('c')
        wf.run()
        self.assertEqual(env.locals['executed'], ['c_0', 'c_1', 'a_2',  'a_2'])
        # allow specifying a single step
        # step will be looped
        script = SoS_Script('''
if 'executed' in locals():
    executed.append(_step.name)
else:
    executed = [_step.name]
[a_1]
[a_2]
[c_0]
[c_1=a_2]
input: 'a.txt', 'b.txt', group_by='single'
''')
        env.run_mode = 'dryrun'
        wf = script.workflow('c')
        wf.run()
        self.assertEqual(env.locals['executed'], ['c_0', 'c_1', 'a_2', 'a_2'])
        #
        # recursive subworkflow not allowed
        script = SoS_Script('''
if 'executed' in locals():
    executed.append(_step.name)
else:
    executed = [_step.name]
[a_1]
[a_2]
[c_0]
[c_1=a_2+c]
input: 'a.txt', 'b.txt', group_by='single'
''')
        env.run_mode = 'dryrun'
        self.assertRaises(RuntimeError, script.workflow, 'c')
        #
        # nested subworkflow is allowed
        script = SoS_Script('''
if 'executed' in locals():
    executed.append(_step.name)
else:
    executed = [_step.name]
[a_1]
[a_2]
[a_3]
[b_1]
[b_2=a_1+a_2]
[c_0]
[c_1=a+b]
input: 'a.txt'
''')
        env.run_mode = 'dryrun'
        wf = script.workflow('c')
        wf.run()
        self.assertEqual(env.locals['executed'], ['c_0', 'c_1', 'a_1', 'a_2', 'a_3',
            'b_1', 'b_2', 'a_1', 'a_2'])
        #
        #
        # nested subworkflow with step option and others
        script = SoS_Script('''
if 'executed' in locals():
    executed.append(_step.name)
else:
    executed = [_step.name]
[a_1]
[a_2]
[a_3]
[b=a_3+a_1, d=a_2, e2_2]
input: 'a.txt', 'b.txt', group_by='single'
''')
        env.run_mode = 'dryrun'
        wf = script.workflow('b')
        wf.run()
        self.assertEqual(env.locals['executed'], ['b_0', 'a_3', 'a_1', 'a_3', 'a_1'])
        wf = script.workflow('d')
        wf.run()
        self.assertEqual(env.locals['executed'], ['d_0', 'a_2', 'a_2'])
        wf = script.workflow('e2')
        wf.run()
        self.assertEqual(env.locals['executed'], ['e2_2'])

    def testSourceOption(self):
        '''Test the source section option'''
        # nested subworkflow with step option and others
        with open('temp/test.sos', 'w') as sos:
            sos.write('''
# test sos script

# global definition
GLB = 5
if 'executed' in locals():
    executed.append('t.' + _step.name)
else:
    executed = ['t.' + _step.name]

[parameters]
parB = 10

[A_1]
output: _input[0] + 'a1'

[A_2]
output: _input[0] + 'a2'

''')
        script = SoS_Script('''
if 'executed' in locals():
    executed.append('g.' + _step.name)
else:
    executed = ['g.' + _step.name]
[b_1=A : source='temp/test.sos', skip=False]
input: 'a.txt', 'b.txt', group_by='single'
''')
        env.run_mode = 'dryrun'
        wf = script.workflow('b')
        wf.run()
        self.assertEqual(env.locals['GLB'], 5)
        self.assertEqual(env.locals['parB'], 10)
        self.assertEqual(env.locals['executed'], ['g.b_1', 't.A_parameters', 't.A_1', 't.A_2', 't.A_parameters', 't.A_1', 't.A_2'])
        #
        os.remove('temp/test.sos')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
#
# This file is part of Script of Scripts (SoS), a workflow system
# for the execution of commands and scripts in different languages.
# Please visit https://github.com/bpeng2000/SOS for more information.
#
# Copyright (C) 2016 Bo Peng (bpeng@mdanderson.org)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

# passing string as unicode to python 2 version of SoS
# to ensure compatibility
from __future__ import unicode_literals

import os
import unittest

from pysos import *

class TestActions(unittest.TestCase):
    def testCheckCommand(self):
        '''Test action check_command'''
        script = SoS_Script(r"""
[0]
check_command('cat')
""")
        wf = script.workflow()
        # should be ok
        wf.run()
        #
        script = SoS_Script(r"""
[0]
check_command('catmouse')
""")
        env.run_mode = 'dryrun'
        wf = script.workflow()
        # should fail in dryrun mode
        self.assertRaises(RuntimeError, wf.run)
        #
        env.run_mode = 'run'
        wf = script.workflow()
        # should fail also in run mode
        self.assertRaises(RuntimeError, wf.run)

    def testFailIf(self):
        '''Test action fail if'''
        script = SoS_Script(r"""
[0]
input: 'a.txt'
fail_if(len(_step.input) == 1)
""")
        env.run_mode = 'dryrun'
        wf = script.workflow()
        # should fail in dryrun mode
        self.assertRaises(RuntimeError, wf.run)
        script = SoS_Script(r"""
[0]
input: 'a.txt', 'b.txt'
fail_if(len(_step.input) == 1)
""")
        env.run_mode = 'dryrun'
        wf = script.workflow()
        # should be ok
        wf.run()

    def testWarnIf(self):
        '''Test action fail if'''
        script = SoS_Script(r"""
[0]
input: 'a.txt'
warn_if(len(_step.input) == 1, 'Expect to see a warning message')
""")
        env.run_mode = 'dryrun'
        wf = script.workflow()
        # should see a warning message.
        wf.run()
        #self.assertRaises(RuntimeError, wf.run)
        script = SoS_Script(r"""
[0]
input: 'a.txt', 'b.txt'
warn_if(len(_step.input) == 1)
""")
        env.run_mode = 'dryrun'
        wf = script.workflow()
        # should be silent
        wf.run()

    def testSearchOutput(self):
        '''Test action search_output'''
        script = SoS_Script(r"""
[0]
search_output('cat test_actions.py', 'abcde' + 'fgh')
""")
        wf = script.workflow()
        # should raise an error
        self.assertRaises(RuntimeError, wf.run)
        #
        script = SoS_Script(r"""
search_output('cat test_actions.py', 'testSearchOutput')
""")
        wf = script.workflow()
        wf.run()

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python2.7
#
# This file is part of Script of Scripts (sos), a workflow system
# for the execution of commands and scripts in different languages.
# Please visit https://github.com/bpeng2000/SOS
#
# Copyright (C) 2016 Bo Peng (bpeng@mdanderson.org)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from distutils.core import setup
from pysos import __version__

setup(name = "sos",
    version = __version__,
    description = "Script of Scripts (SoS): a workflow system for the execution of scripts in different languages",
    author = 'Bo Peng',
    url = 'https://github.com/bpeng2000/SOS',
    author_email = 'bpeng@mdanderson.org',
    maintainer = 'Bo Peng',
    maintainer_email = 'bpeng@mdanderson.org',
    license = 'GPL3',
    classifiers = [
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS :: MacOS X',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3 :: Only',
        ],
    packages = ['pysos'],
    install_requires=[
          'psutil',
      ],
    scripts = [
        'sos',
        'sos-runner',
        ],
    py_modules = [
        'pysos.actions',
        'pysos.utils',
        'pysos.sos_script',
        ]
)



#!/usr/bin/env python3
#
# This file is part of Script of Scripts (sos), a workflow system
# for the execution of commands and scripts in different languages.
# Please visit https://github.com/bpeng2000/SOS
#
# Copyright (C) 2016 Bo Peng (bpeng@mdanderson.org)
##
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import os
import re
import subprocess
import tempfile
import pipes
from .utils import env

#
# A decoration function that allows SoS to replace all SoS actions
# with a null action.
#
def SoS_Action(run_mode='run'):
    run_mode = [run_mode] if isinstance(run_mode, str) else run_mode
    def runtime_decorator(func):
        def action_wrapper(*args, **kwargs):
            if env.run_mode not in run_mode:
                return 0
            else:
                return func(*args, **kwargs)
        return action_wrapper
    return runtime_decorator


class SoS_ExecuteScript:
    def __init__(self, script, interpreter, suffix):
        self.script = script
        self.interpreter = interpreter
        self.script_file = tempfile.NamedTemporaryFile(mode='w+t', suffix=suffix, delete=False).name
        with open(self.script_file, 'w') as script_file:
            script_file.write(self.script)

    def run(self):
        if '{}' in self.interpreter:
            cmd = self.interpreter.replace('{}', pipes.quote(self.script_file))
        else:
            cmd = self.interpreter + ' ' + pipes.quote(self.script_file)
        try:
            p = subprocess.Popen(cmd, shell=True)
            env.register_process(p.pid, 'Runing {}'.format(self.script_file))
            ret = p.wait()
        finally:
            env.deregister_process(p.pid)
        if ret != 0:
            raise RuntimeError('Failed to execute script')



class SoS_ExecuteRScript(SoS_ExecuteScript):
    '''SoS_Execute in-line R script using Rscript as interpreter. Please
    check action SoS_ExecuteScript for more details.
    '''
    def __init__(self, script=''):
        SoS_ExecuteScript.__init__(self, script=script, interpreter='Rscript',
            suffix='.R')

class SoS_ExecuteShellScript(SoS_ExecuteScript):
    '''SoS_Execute in-line shell script using bash as interpreter. Please
    check action SoS_ExecuteScript for more details.
    '''
    def __init__(self, script=''):
        SoS_ExecuteScript.__init__(self, script=script, interpreter='bash',
            suffix='.sh')

class SoS_ExecutePythonScript(SoS_ExecuteScript):
    '''SoS_Execute in-line python script using python as interpreter. Please
    check action SoS_ExecuteScript for more details.
    '''
    def __init__(self, script=''):
        SoS_ExecuteScript.__init__(self, script=script, interpreter='python',
            suffix='.py')


class SoS_ExecutePython3Script(SoS_ExecuteScript):
    '''SoS_Execute in-line python script using python as interpreter. Please
    check action SoS_ExecuteScript for more details.
    '''
    def __init__(self, script=''):
        SoS_ExecuteScript.__init__(self, script=script, interpreter='python3',
            suffix='.py')

@SoS_Action(run_mode='run')
def run(script):
    SoS_ExecuteShellScript(script).run()

@SoS_Action(run_mode='run')
def python(script):
    SoS_ExecutePythonScript(script).run()

@SoS_Action(run_mode='run')
def python3(script):
    SoS_ExecutePython3Script(script).run()

@SoS_Action(run_mode='run')
def R(script):
    SoS_ExecuteRScript(script).run()


try:
    from shutil import which
except ImportError:
    # this function is only define in Python 3.3 +
    def which(cmd, mode=os.F_OK | os.X_OK, path=None):
        """Given a command, mode, and a PATH string, return the path which
        conforms to the given mode on the PATH, or None if there is no such
        file.

        `mode` defaults to os.F_OK | os.X_OK. `path` defaults to the result
        of os.environ.get("PATH"), or can be overridden with a custom search
        path.

        """
        # Check that a given file can be accessed with the correct mode.
        # Additionally check that `file` is not a directory, as on Windows
        # directories pass the os.access check.
        def _access_check(fn, mode):
            return (os.path.exists(fn) and os.access(fn, mode)
                    and not os.path.isdir(fn))

        # Short circuit. If we're given a full path which matches the mode
        # and it exists, we're done here.
        if _access_check(cmd, mode):
            return cmd

        path = (path or os.environ.get("PATH", os.defpath)).split(os.pathsep)
        files = [cmd]

        seen = set()
        for dir in path:
            dir = os.path.normcase(dir)
            if not dir in seen:
                seen.add(dir)
                for thefile in files:
                    name = os.path.join(dir, thefile)
                    if _access_check(name, mode):
                        return name
        return None


@SoS_Action(run_mode=['dryrun', 'run'])
def check_command(cmds):
    '''Check the existence of command `cmd` and raise an error if
    command does not exist. `cmd` can be one command or a list of
    commands.'''
    cmds = [cmds] if isinstance(cmds, str) else cmds
    #
    for cmd in cmds:
        name = which(cmd)
        if not name:
            raise RuntimeError('Command ``{}`` not found!'.format(cmd))
        env.logger.info('Command ``{}`` is located as ``{}``.'.format(cmd, name))
    return 0


@SoS_Action(run_mode=['dryrun', 'run'])
def fail_if(expr, msg=''):
    '''Raise an exception with `msg` if condition `expr` is False'''
    if expr:
        raise RuntimeError(msg)


@SoS_Action(run_mode=['dryrun', 'run'])
def warn_if(expr, msg=''):
    '''Yield an warning message `msg` if `expr` is False '''
    if expr:
        env.logger.warning(msg)

@SoS_Action(run_mode=['dryrun', 'run'])
def search_output(cmd, pattern):
    '''Raise an exception if output of `cmd` does not match specified `pattern`.
    Multiple patterns can be specified as a list of patterns.'''
    output = subprocess.check_output(cmd, stderr=subprocess.STDOUT, shell=True).decode()
    env.logger.trace('Output of command ``{}`` is ``{}``'.format(cmd, output))
    #
    pattern = [pattern] if isinstance(pattern, str) else pattern
    if all([re.search(x, output, re.MULTILINE) is None for x in pattern]):
        raise RuntimeError('Output of command ``{}`` does not match specified regular expression ``{}``.'
            .format(cmd, ' or '.join(pattern)))
